/***************************************************************************
 *   Copyright (C) 2006 by Kevin Gilbert                                   *
 *   kev.gilbert@cdu.edu.au                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <math.h>
#include <time.h>

#include <qdatetime.h>
#include <qstringlist.h>
#include <qtimer.h>
#include <qwidgetstack.h>

#include <kapplication.h>
#include <kconfig.h>
#include <kdebug.h>
#include <kglobal.h>
#include <kglobalsettings.h>
#include <klocale.h>
#include <kmessagebox.h>

#include "authenticate.h"
#include "clientnamedlg.h"
#include "disconnecteddlg.h"
#include "gui_rpc_client.h"
#include "mainwidget.h"
#include "mainwidgetcontrol.h"
#include "mainwidgetview.h"

MainWidget* MainWidget::m_self;

//	constructor
//	===========

MainWidget::MainWidget( QWidget* parent, const char* name )
: QSplitter( parent, name ),
  m_currentClient( NULL ),
  m_disconnectedCancelled( false ),
  m_disconnectedDlg( NULL ),
  m_firstUpdate( true ),
  m_messageTimer( new QTimer( this, "message timer" )),
  m_updateTimer( new QTimer( this, "update timer" )),
  m_wallet( NULL )
{	m_clients.setAutoDelete( true );
	m_filters.setAutoDelete( true );
	m_filters.append( new MessageFilter( ));
	m_self = this;

	m_controlView = new MainWidgetControl( this, "main control" );
	m_stackView   = new QWidgetStack( this, "main stack" );

	connect( m_controlView,
			 SIGNAL( changeCurrentClient( const CPtr )),
			 SLOT( slotChangeCurrentClient( const CPtr )));

	connect( m_messageTimer,
			 SIGNAL( timeout( )),
			 SLOT( slotUpdateMessage( )));

	connect( m_updateTimer,
			 SIGNAL( timeout( )),
			 SLOT( slotUpdate( )));

	m_updateTimer->start( 0, true );
}

//	closeWallet
//	===========

void MainWidget::closeWallet( )
{	if( m_wallet == NULL )
		return;

	Wallet::closeWallet( m_walletName, false );
	delete m_wallet;
	m_wallet = NULL;
}

//	connectToClient
//	===============

CPtr MainWidget::connectToClient( const QString& clientName, const ushort portNo, const QString& rpcAuth )
{	CPtr client;

	for( client = m_clients.first( ); client != NULL; client = m_clients.next( ))
		if( client->clientName == clientName )
		{	client->portNo  = portNo;
			client->rpcAuth = rpcAuth;
			return connectToClient( client ) ? client : NULL;
		}

	client               = new Client;
	client->clientName   = clientName;
	client->authenticate = new Authenticate( client, m_wallet, this, "authenticate" );
	client->locked       = false;
	client->portNo       = portNo;
	client->rpcAuth      = rpcAuth;

	m_clients.append( client );

	if( connectToClient( client ))
		return client;

	client->retryCount = 0;
	client->retryTime  = 0;
	updateDisconnectedClientStatus( client, i18n( "Retry on next update cycle" ));
	updateDisconnectedClientRetryCount( client );

	return NULL;
}

//	connectToClient
//	===============

bool MainWidget::connectToClient( const CPtr client )
{	int statr = client->rpcClient.init( client->clientName, client->portNo );

	if( statr )
	{	client->reason.sprintf( i18n( "rpc init failed with status %d" ), statr );
		connectToClientFailed( client );
		return false;
	}


	if( !client->rpcAuth.isEmpty( ))
	{	statr = client->rpcClient.authorize( client->rpcAuth );

		if( statr )
		{	client->reason.sprintf( i18n( "rpc authorisation failed with status %d" ), statr );
			connectToClientFailed( client );
			return false;
		}
	}

	PROJECTS projects;
	statr = client->rpcClient.get_project_status( projects );

	if( statr )
	{	client->reason.sprintf( i18n( "get_project_status returned error %d" ), statr );
		connectToClientFailed( client );
		return false;
	}

	ushort size = projects.projects.size( );

	if( !size )
	{	client->reason = "no projects - probable connection failure";
		connectToClientFailed( client );
		return false;
	}

	client->connected = true;
	client->listView  = new MainWidgetView( client, m_font, this, "main view" );
	client->reason    = QString::null;

	m_controlView->addClient( client );
	m_stackView->addWidget( client->listView );
	client->listView->show( );

	connect( client->listView, SIGNAL( hasKonsole( const bool )), SIGNAL( hasKonsole( const bool )));
	connect( this, SIGNAL( walletChanged( Wallet* )), client->authenticate, SLOT( slotWalletChanged( Wallet* )));

	return true;
}

//	connectToClientFailed
//	=====================

void MainWidget::connectToClientFailed( const CPtr client )
{	client->connected = false;
	client->listView  = NULL;
	client->manual    = false;
}

//	createDisconnectedClientDlg
//	===========================

bool MainWidget::createDisconnectedClientDlg( )
{	if( m_disconnectedCancelled )
		return false;

	if( m_disconnectedDlg != NULL )
		return true;

	m_disconnectedDlg = new DisconnectedDlg( this, "disconnected dlg" );
	m_disconnectedDlg->show( );

	connect( m_disconnectedDlg,
			 SIGNAL( cancelClicked( )),
			 SLOT( slotDisconnectedDlgCancelled( )));

	connect( m_disconnectedDlg,
			 SIGNAL( retryConnection( const CPtr )),
			 SLOT( slotRetryConnection( const CPtr )));

	return true;
}

//	disconnectClient
//	================

CPtr MainWidget::disconnectClient( const QString& clientName )
{	CPtr client;

	for( client = m_clients.first( ); client != NULL; client = m_clients.next( ))
		if( client->clientName == clientName )
			break;

	if( client == NULL )
	{	Q_ASSERT( false );
		return NULL;
	}

	disconnectClient( client );
	client->manual = true;
	return client;
}

//	disconnectClient
//	================

void MainWidget::disconnectClient( const CPtr client )
{	client->rpcClient.close( );
	client->connected = false;

	if( client->reason == QString::null )
		client->reason = i18n( "manual disconnect" );

	m_controlView->removeClient( client );
	m_stackView->removeWidget( client->listView );
	delete client->listView;
	client->listView = NULL;

	if( client != m_currentClient )
		return;

	for( m_currentClient = m_clients.first( ); m_currentClient != NULL; m_currentClient = m_clients.next( ))
		if( m_currentClient->connected )
			break;

	slotChangeCurrentClient( m_currentClient );
}

//	filtersUpdated
//	==============

void MainWidget::filtersUpdated( )
{	saveFilters( );

	for( CPtr client = m_clients.first( ); client != NULL; client = m_clients.next( ))
		if( client->listView != NULL )
			client->listView->filtersUpdated( );
}

//	firstUpdate
//	===========

void MainWidget::firstUpdate( )
{	if( m_firstUpdate )
	{	m_firstUpdate = false;
		m_updateTimer->start( m_updateInterval );
	}

	slotChangeCurrentClient( m_currentClient );
}

//	openWallet
//	==========

Wallet* MainWidget::openWallet( )
{	if( m_wallet != NULL )
		return NULL;

	m_wallet = Wallet::openWallet( m_walletName );

	if( m_wallet == NULL )
	{	KMessageBox::sorry( NULL, i18n( QString( "Could not open the \"%1\" wallet. No authentication data will be saved!" ).arg( m_walletName )), i18n( "Wallet open failed" ));
		return NULL;
	}

	if( m_wallet->setFolder( Wallet::PasswordFolder( )))
		return m_wallet;

	if( !m_wallet->createFolder( Wallet::PasswordFolder( )))
	{	closeWallet( );
		return NULL;
	}

	if( m_wallet->setFolder( Wallet::PasswordFolder( )))
		return m_wallet;

	KMessageBox::sorry( NULL, i18n( QString( "Could not open the \"Passwords\" folder in the \"%1\" wallet. No authentication data will be saved!" ).arg( m_walletName )), i18n( "Wallet folder open failed" ));
	return NULL;
}

//	readFilters
//	===========

void MainWidget::readFilters( )
{	KConfig* config = kapp->config( );
	config->setGroup( "Filters" );

	ushort noOfFilters = config->readNumEntry( "noOfFilters" );
	ushort index;

	for( index = 0; index< noOfFilters; index++ )
		m_filters.append( new MessageFilter( config->readEntry( "filter" + QString::number( index ))));
}

//	readSettings
//	============

void MainWidget::readSettings( )
{	KConfig* config = kapp->config( );
	config->setGroup( "MainWindow" );

	QColor defaultColour     = KGlobalSettings::textColor( );

	m_projectUpdateInterval           = config->readNumEntry( "projectUpdateInterval", 1 );
	m_projectUpdateTOD                = config->readNumEntry( "projectUpdateTOD" );
	m_retryExpBack                    = config->readBoolEntry( "retryExpBack", true );
	m_retryExpBackLimit               = config->readNumEntry( "retryExpBackLimit", 600 );
	m_retryLimit                      = config->readNumEntry( "retryLimit", -1 );
	m_retryLostCxns                   = config->readBoolEntry( "retryLostCxns", true );
	m_standardFilter.alertErrorAction = FilterAction( config->readNumEntry( "stdFilterAlertErrorAction" ));
	m_standardFilter.alertErrorColour = config->readColorEntry( "stdFilterAlertErrorColour", &defaultColour );
	m_standardFilter.alertInfoAction  = FilterAction( config->readNumEntry( "stdFilterAlertInfoAction" ));
	m_standardFilter.alertInfoColour  = config->readColorEntry( "stdFilterAlertInfoColour", &defaultColour );
	m_standardFilter.errorAction      = FilterAction( config->readNumEntry( "stdFilterErrorAction" ));
	m_standardFilter.errorColour      = config->readColorEntry( "stdFilterErrorColour", &defaultColour );
	m_standardFilter.infoAction       = FilterAction( config->readNumEntry( "stdFilterInfoAction" ));
	m_standardFilter.infoColour       = config->readColorEntry( "stdFilterInfoColour", &defaultColour );
	m_standardFilter.userAction       = FilterAction( config->readNumEntry( "stdFilterUserAction" ));
	m_standardFilter.userColour       = config->readColorEntry( "stdFilterUserColour", &defaultColour );
	m_synchroniseLayouts              = config->readBoolEntry( "synchroniseLayouts", false );
	m_updateInterval                  = config->readNumEntry( "updateInterval", 5000 );
	m_walletName                      = config->readEntry( "walletName", "kdewallet" );

	QTime projectUpdateTOD( m_projectUpdateTOD / 3600, (m_projectUpdateTOD / 60) % 60, m_projectUpdateTOD % 60 );
	m_projectTimeNext = QDateTime::currentDateTime( );
	m_projectTimeNext.setTime( projectUpdateTOD );

	setSizes( config->readIntListEntry( "viewWidths" ));
	m_font.fromString( config->readEntry( "font" ));
	QStringList clientList = config->readListEntry( "clients" );

	for( QStringList::Iterator it = clientList.begin( ); it != clientList.end( ); ++it )
	{	QStringList clientDets = QStringList::split( "\034", *it );
		connectToClient( clientDets[ 0 ], clientDets[ 1 ].toUShort( ), clientDets[ 2 ]);
	}

	config->setGroup( "MainWindow" );
	QString currentClientName = config->readEntry( "currentClientName" );

	if( currentClientName == QString::null )
		m_currentClient = m_clients.first( );
	else
	{	CPtr client;

		for( client = m_clients.first( ); client != NULL; client = m_clients.next( ))
			if( client->clientName == currentClientName )
				break;

		m_currentClient = client == NULL ? m_clients.first( ) : client;
	}

	readFilters( );
}

//	removeDisconnectedClient
//	========================

void MainWidget::removeDisconnectedClient( const CPtr client )
{	if( m_disconnectedDlg == NULL )
		return;

	bool disconnectedCancelled = m_disconnectedCancelled;
	m_disconnectedDlg->removeClient( client );
	m_disconnectedCancelled = disconnectedCancelled;
}

//	retryConnection
//	===============

bool MainWidget::retryConnection( const CPtr client, const time_t now )
{	if( !m_retryLostCxns )
	{	updateDisconnectedClientStatus( client, i18n( "Not configured to retry connections" ));
		return false;
	}

	if( m_retryLimit > 0 && client->retryCount >= m_retryLimit )
	{	updateDisconnectedClientStatus( client, i18n( "Retry limit exceeded" ));
		return false;
	}

	if( m_retryExpBack && client->retryTime > now )
	{	QDateTime retryTime;
		retryTime.setTime_t( client->retryTime );
		updateDisconnectedClientStatus( client, i18n( "Retry on first update after " ) + KGlobal::locale( )->formatDateTime( retryTime, true, true ));
		return false;
	}

	if( connectToClient( client ))
	{	removeDisconnectedClient( client );
		client->reason = "";
		slotChangeCurrentClient( client );
		return true;
	}

	client->retryCount++;
	updateDisconnectedClientRetryCount( client );

	if( !m_retryExpBack )
	{	QDateTime retryTime;
		retryTime.setTime_t( now + m_updateInterval );
		updateDisconnectedClientStatus( client, i18n( "Retry on first update after " ) + KGlobal::locale( )->formatDateTime( retryTime, true, true ));
		return false;
	}

	double interval = pow( 2, client->retryCount - 1 );

	if( interval > m_retryExpBackLimit )
		interval = m_retryExpBackLimit;

	client->retryTime = now + uint( interval );

	QDateTime retryTime;
	retryTime.setTime_t( client->retryTime );
	updateDisconnectedClientStatus( client, i18n( "Retry on first update after " ) + KGlobal::locale( )->formatDateTime( retryTime, true, true ));

	return false;
}

//	saveFilters
//	===========

void MainWidget::saveFilters( )
{	KConfig* config = kapp->config( );
	config->deleteGroup( "Filters" );
	config->setGroup( "Filters" );
	config->writeEntry( "noOfFilters", m_filters.count( ) - 1 );
	ushort index = 0;

	for( MFPtr filter = m_filters.first( ); filter != NULL; filter = m_filters.next( ))
		if( !filter->isStandardFilter( ))
			config->writeEntry( "filter" + QString::number( index++ ), filter->toString( ));

	config->sync( );
}

//	saveSettings
//	============

void MainWidget::saveSettings( )
{	for( CPtr client = m_clients.first( ); client != NULL; client = m_clients.next( ))
		if( client->connected )
			client->listView->saveSettings( true );

	if( m_synchroniseLayouts && m_clients.first( ) != NULL && m_clients.first( )->connected )
		m_clients.first( )->listView->saveSettings( false );

	KConfig* config = kapp->config( );
	config->setGroup( "MainWindow" );
	config->writeEntry( "font",                      m_font.toString( ));
	config->writeEntry( "projectUpdateInterval",     m_projectUpdateInterval );
	config->writeEntry( "projectUpdateTOD",          m_projectUpdateTOD );
	config->writeEntry( "retryExpBack",              m_retryExpBack );
	config->writeEntry( "retryExpBackLimit",         m_retryExpBackLimit );
	config->writeEntry( "retryLimit",                m_retryLimit );
	config->writeEntry( "retryLostCxns",             m_retryLostCxns );
	config->writeEntry( "stdFilterAlertErrorAction", m_standardFilter.alertErrorAction );
	config->writeEntry( "stdFilterAlertErrorColour", m_standardFilter.alertErrorColour );
	config->writeEntry( "stdFilterAlertInfoAction",  m_standardFilter.alertInfoAction );
	config->writeEntry( "stdFilterAlertInfoColour",  m_standardFilter.alertInfoColour );
	config->writeEntry( "stdFilterErrorAction",      m_standardFilter.errorAction );
	config->writeEntry( "stdFilterErrorColour",      m_standardFilter.errorColour );
	config->writeEntry( "stdFilterInfoAction",       m_standardFilter.infoAction );
	config->writeEntry( "stdFilterInfoColour",       m_standardFilter.infoColour );
	config->writeEntry( "stdFilterUserAction",       m_standardFilter.userAction );
	config->writeEntry( "stdFilterUserColour",       m_standardFilter.userColour );
	config->writeEntry( "synchroniseLayouts",        m_synchroniseLayouts );
	config->writeEntry( "updateInterval",            m_updateInterval );
	config->writeEntry( "viewWidths",                sizes( ));
	config->writeEntry( "walletName",                m_walletName );

	if( m_currentClient != NULL )
		config->writeEntry( "currentClientName", m_currentClient->clientName );

	CPtr client;
	QStringList clientList;

	for( client = m_clients.first( ); client != NULL; client = m_clients.next( ))
		clientList.append( client->clientName + "\034" + QString::number( client->portNo ) + "\034" + client->rpcAuth );

	config->writeEntry( "clients", clientList );
	saveFilters( );
	config->sync( );
}

//	showEvent
//	=========

void MainWidget::showEvent( QShowEvent* event )
{	QSplitter::showEvent( event );
	slotChangeCurrentClient( m_currentClient );
}

//	slotApplyPreferences
//	====================

void MainWidget::slotApplyPreferences( const Preferences::PrefData* prefData )
{	m_projectUpdateInterval = prefData->projectUpdateInterval;
	m_projectUpdateTOD      = prefData->projectUpdateTOD;
	m_retryExpBack          = prefData->retryExpBack;
	m_retryExpBackLimit     = prefData->retryExpBackLimit;
	m_retryLimit            = prefData->retryLimit;
	m_retryLostCxns         = prefData->retryLostCxns;
	m_synchroniseLayouts    = prefData->synchroniseLayouts;
	m_updateInterval        = prefData->updateInterval;

	m_updateTimer->changeInterval( m_updateInterval );
	slotUpdateMessage( );

	if( prefData->fontUpdated )
	{	m_font = prefData->font;

		for( CPtr client = m_clients.first( ); client != NULL; client = m_clients.next( ))
			if( client->listView != NULL )
				client->listView->changeFont( m_font );
	}

	if( m_walletName != prefData->walletName )
	{	m_walletName = prefData->walletName;
		closeWallet( );
		openWallet( );
		emit( walletChanged( m_wallet ));
	}

	saveSettings( );
}

//	slotChangeCurrentClient
//	=======================

void MainWidget::slotChangeCurrentClient( const CPtr currentClient )
{	m_currentClient = currentClient;

	if( m_currentClient == NULL || m_currentClient->listView == NULL )
	{	emit( changeCurrentClient( "no current client" ));
		return;
	}

	m_stackView->raiseWidget( m_currentClient->listView );
	m_controlView->setCurrentClient( m_currentClient );
	emit( changeCurrentClient( m_currentClient->clientName + ": " + m_currentClient->listView->toolTipText( )));
}

//	slotConnectClient
//	=================

void MainWidget::slotConnectClient( )
{	ClientNameDlg dlg( m_clients, ClientNameDlg::Connect, i18n( "Connect to client" ), this, "client name dlg" );

	if( dlg.exec( ) != QDialog::Accepted )
		return;

	CPtr client = connectToClient( dlg.clientName( ), dlg.portNo( ), dlg.rpcAuth( ));

	if( client == NULL )
		return;

	slotChangeCurrentClient( client );
	slotUpdate( );
	m_disconnectedCancelled = false;
}

//	slotDeleteClient
//	================

void MainWidget::slotDeleteClient( )
{	ClientNameDlg dlg( m_clients, ClientNameDlg::Delete, i18n( "Delete client" ), this, "client name dlg" );

	if( dlg.exec( ) != QDialog::Accepted )
		return;

	CPtr client = disconnectClient( dlg.clientName( ));

	if( client != NULL )
		m_clients.removeRef( client );
}

//	slotDisconnectClient
//	====================

void MainWidget::slotDisconnectClient( )
{	ClientNameDlg dlg( m_clients, ClientNameDlg::Disconnect, i18n( "Disconnect from client" ), this, "client name dlg" );

	if( dlg.exec( ) == QDialog::Accepted )
		disconnectClient( dlg.clientName( ));
}

//	slotDisconnectedDlgCancelled
//	============================

void MainWidget::slotDisconnectedDlgCancelled( )
{	m_disconnectedCancelled = true;
	m_disconnectedDlg = NULL;
}

//	slotKonsoleClose
//	================

void MainWidget::slotKonsoleClose( )
{	m_currentClient->listView->konsoleClose( );
}

//	slotKonsoleOpen
//	===============

void MainWidget::slotKonsoleOpen( )
{	m_currentClient->listView->konsoleOpen( );
}

//	slotPreferences
//	===============

void MainWidget::slotPreferences( )
{	Preferences::PrefData* prefData = new Preferences::PrefData;
	prefData->font                  = m_font;
	prefData->projectUpdateInterval = m_projectUpdateInterval;
	prefData->projectUpdateTOD      = m_projectUpdateTOD;
	prefData->retryExpBack          = m_retryExpBack;
	prefData->retryExpBackLimit     = m_retryExpBackLimit;
	prefData->retryLimit            = m_retryLimit;
	prefData->retryLostCxns         = m_retryLostCxns;
	prefData->synchroniseLayouts    = m_synchroniseLayouts;
	prefData->updateInterval        = m_updateInterval;
	prefData->walletName            = m_walletName;

	Preferences dlg( prefData, this, "preferences dlg" );
	connect( &dlg,
			 SIGNAL( applyPreferences( const Preferences::PrefData* )),
			 SLOT( slotApplyPreferences( const Preferences::PrefData* )));
	dlg.exec( );
}

//	slotRetryConnection
//	===================

void MainWidget::slotRetryConnection( const CPtr client )
{	client->retryCount = 0;
	client->retryTime  = 0;

	if( retryConnection( client, time( NULL )))
		slotUpdate( );
}

//	slotUpdate
//	==========

void MainWidget::slotUpdate( )
{	update( );
}

//	slotUpdateMessage
//	=================

void MainWidget::slotUpdateMessage( )
{	QString projectTimeNextText;

	if( !m_projectUpdateInterval )
		projectTimeNextText = "(never)";
	else
	{	QDateTime projectTimeNext = m_projectTimeNext < m_updateTimeNext ? m_updateTimeNext : m_projectTimeNext;
		projectTimeNextText = KGlobal::locale( )->formatDateTime( projectTimeNext, true, true );
	}

	QString   text;
	text.sprintf( "Last update: %s, next update due: %s, next Projects Update: %s",
				  KGlobal::locale( )->formatDateTime( m_updateTimePrior, true, true ).ascii( ),
				  KGlobal::locale( )->formatDateTime( m_updateTimeNext,  true, true ).ascii( ),
				  projectTimeNextText.ascii( ));

	emit( statusBarMessage( text ));
}

//	slotUpdateNow
//	=============

void MainWidget::slotUpdateNow( )
{	update( );
	m_updateTimer->start( m_updateInterval );
}

//	standardFilter
//	==============

void MainWidget::standardFilter( const StdFilter s )
{	m_standardFilter.alertErrorAction = s.alertErrorAction;
	m_standardFilter.alertErrorColour = s.alertErrorColour;
	m_standardFilter.alertInfoAction  = s.alertInfoAction;
	m_standardFilter.alertInfoColour  = s.alertInfoColour;
	m_standardFilter.errorAction      = s.errorAction;
	m_standardFilter.errorColour      = s.errorColour;
	m_standardFilter.infoAction       = s.infoAction;
	m_standardFilter.infoColour       = s.infoColour;
	m_standardFilter.userAction       = s.userAction;
	m_standardFilter.userColour       = s.userColour;

	saveFilters( );

	for( CPtr client = m_clients.first( ); client != NULL; client = m_clients.next( ))
		if( client->listView != NULL )
			client->listView->filtersUpdated( );
}

//	toolTipText
//	===========

QString MainWidget::toolTipText( )
{	bool    first = true;
	QString text;

	for( CPtr client = m_clients.first( ); client != NULL; client = m_clients.next( ))
	{	if( first)
			first = false;
		else
			text += '\n';

		text += client->clientName + ": ";

		if( !client->connected )
			text += "disconnected";
		else
			text += client->listView->toolTipText( );
	}

	return text;
}

//	update
//	======

void MainWidget::update( )
{	bool   didSomething = false;
	ushort disconnected = 0;
	time_t now          = time( NULL );
	QPtrList<Client> badClients;

	for( CPtr client = m_clients.first( ); client != NULL; client = m_clients.next( ))
	{	if( !client->connected )
		{	disconnected++;

			if( client->manual || !retryConnection( client, now ))
				continue;
		}

		emit( statusBarMessage( "Updating " + client->clientName ));

		if( client->listView->update( ))
			didSomething = true;
		else
		{	disconnected++;
			badClients.append( client );
		}
	}

	QString text;

	if( didSomething )
		text = "All hosts updated!";
	else
		text = "No active hosts!";

	if( disconnected == 1 )
		text += " (There is 1 disconnected client)";
	else
	if( disconnected )
		text += QString( " (There are %1 disconnected clients)" ).arg( disconnected );

	if( m_projectUpdateInterval && m_projectTimeNext < QDateTime::currentDateTime( ))
		updateProjects( );

	emit( statusBarMessage( text ));
	m_updateTimePrior = QDateTime::currentDateTime( );
	m_updateTimeNext  = m_updateTimePrior.addSecs( m_updateInterval / 1000 );
	m_messageTimer->start( 1000, true );

	if( !badClients.count( ))
	{	firstUpdate( );
		return;
	}

	for( CPtr client = badClients.first( ); client != NULL; client = badClients.next( ))
		if( !client->locked )
		{	disconnectClient( client );

			client->manual     = false;
			client->retryCount = 0;
			client->retryTime  = 0;

			updateDisconnectedClientStatus( client, i18n( "Retry on next update cycle" ));
		}

	firstUpdate( );
}

//	updateDisconnectedClientStatus
//	==============================

void MainWidget::updateDisconnectedClientStatus( const CPtr client, const QString& text )
{	if( createDisconnectedClientDlg( ))
		m_disconnectedDlg->updateStatus( client, text );
}

//	updateDisconnectedClientRetryCount
//	==================================

void MainWidget::updateDisconnectedClientRetryCount( const CPtr client )
{	if( createDisconnectedClientDlg( ))
		m_disconnectedDlg->updateRetryCount( client );
}

//	updateProjects
//	==============

void MainWidget::updateProjects( )
{	QTime projectUpdateTOD( m_projectUpdateTOD / 3600, (m_projectUpdateTOD / 60) % 60, m_projectUpdateTOD % 60 );
	m_projectTimeNext = QDateTime::currentDateTime( ).addSecs( 86400 * m_projectUpdateInterval );
	m_projectTimeNext.setTime( projectUpdateTOD );

	for( CPtr client = m_clients.first( ); client != NULL; client = m_clients.next( ))
		if( client->connected )
			client->listView->updateProjects( );
}

//	wrapUp
//	======

void MainWidget::wrapUp( )
{	closeWallet( );

	delete m_messageTimer;
	delete m_updateTimer;
}
