/***************************************************************************
 *   Copyright (C) 2006 by Kevin Gilbert                                   *
 *   kev.gilbert@cdu.edu.au                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <qdir.h>
#include <qfont.h>
#include <qlayout.h>
#include <qstrlist.h>

#include <kapplication.h>
#include <kconfig.h>
#include <kde_terminal_interface.h>
#include <kdebug.h>
#include <klibloader.h>
#include <klocale.h>
#include <kparts/part.h>

#include "authenticate.h"
#include "mainwidgetview.h"
#include "messagesview.h"
#include "projectsview.h"
#include "resultsview.h"
#include "transfersview.h"
#include "statisticsview.h"

//	constructor
//	===========

MainWidgetView::MainWidgetView( const CPtr   client,
								const QFont& font,
								QWidget*     parent,
								const char*  name )
: KTabWidget( parent, name ),
  m_client( client ),
  m_haveStateData( false ),
  m_konsoleWidget( NULL ),
  m_messagesView( new MessagesView( client, this, "messages view" )),
  m_projectsView( new ProjectsView( client, this, "projects view" )),
  m_resultsView( new ResultsView( client, this, "results view" )),
  m_statisticsView( new StatisticsView( client, this, "statistics view" )),
  m_transfersView( new TransfersView( client, this, "transfers view" ))
{	KConfig* config = kapp->config( );
	config->setGroup( "General_" + m_client->clientName );
	m_hasKonsole = config->readBoolEntry( "hasKonsole", true );
	createKonsolePart( );

	insertTab( m_projectsView,   "Projects"   );
	insertTab( m_resultsView,    "Results"    );
	insertTab( m_messagesView,   "Messages"   );
	insertTab( m_transfersView,  "Transfers"  );
	insertTab( m_statisticsView, "Statistics" );

	if( m_konsoleWidget != NULL )
		insertTab( m_konsoleWidget, "Konsole" );

	changeFont( font );

	connect( m_client->authenticate,
			 SIGNAL( authenticationFinished( const CPtr, const bool )),
			 m_transfersView,
			 SLOT( slotAuthenticationFinished( const CPtr, const bool )));

	connect( m_client->authenticate,
			 SIGNAL( authenticationFinished( const CPtr, const bool )),
			 m_messagesView,
			 SLOT( slotAuthenticationFinished( const CPtr, const bool )));

	connect( m_client->authenticate,
			 SIGNAL( authenticationStarted( const CPtr )),
			 m_messagesView,
			 SLOT( slotAuthenticationStarted( const CPtr )));

	connect( m_client->authenticate,
			 SIGNAL( disableAuthButton( )),
			 m_transfersView,
			 SLOT( slotDisableAuthButton( )));

	connect( m_client->authenticate,
			 SIGNAL( enableAuthButton( )),
			 m_transfersView,
			 SLOT( slotEnableAuthButton( )));

	connect( m_messagesView,
			 SIGNAL( doProjectsUpdate( )),
			 m_projectsView,
			 SLOT( slotUpdateProjects( )));

	connect( m_projectsView,
			 SIGNAL( autoProjectUpdateStarted( )),
			 m_messagesView,
			 SLOT( slotAutoProjectUpdateStarted( )));
}

//	changeFont
//	==========

void MainWidgetView::changeFont( const QFont& font )
{	m_messagesView->setFont( font );
	m_projectsView->setFont( font );
	m_resultsView->setFont( font );
	m_statisticsView->setFont( font );
	m_transfersView->setFont( font );
}

//	createKonsolePart
//	=================

void MainWidgetView::createKonsolePart( )
{	if( !m_hasKonsole )
		return;

	KLibFactory* factory = KLibLoader::self( )->factory( "libkonsolepart" );

	if( factory == NULL )
		return;

	m_konsoleWidget = new QWidget( this, "konsole widget" );
	KParts::ReadOnlyPart* konsolePart = (KParts::ReadOnlyPart*) factory->create( m_konsoleWidget,
																				"knosole part",
																				"QObject",
																				"KParts::ReadOnlyPart" );
	if( konsolePart == NULL )
	{	delete m_konsoleWidget;
		m_konsoleWidget = NULL;
		return;
	}

	konsolePart->widget( )->setFocusPolicy( QWidget::WheelFocus );
	setFocusProxy( konsolePart->widget( ));
	konsolePart->widget( )->setFocus( );

	QVBoxLayout* vbox = new QVBoxLayout( m_konsoleWidget );
	vbox->addWidget( konsolePart->widget( ));
	m_konsoleWidget->show( );

	m_termInterface = static_cast<TerminalInterface*> (konsolePart->qt_cast( "TerminalInterface" ));

	if( m_termInterface != NULL )
	{	m_termInterface->showShellInDir( QDir::home( ).path( ));
	}

	connect( konsolePart,
			 SIGNAL( destroyed( )),
			 SLOT( slotKonsolePartDestroyed( )));
}

//	filtersUpdated
//	==============

void MainWidgetView::filtersUpdated( )
{	m_messagesView->filtersUpdated( );
}

//	getState
//	========

bool MainWidgetView::getState( const CPtr client )
{	int statr = client->rpcClient.get_state( client->state );

	if( !statr )
		return true;

	client->reason.sprintf( i18n( "couldn't get state data, error %d" ), statr );
	return false;
}

//	konsoleClose
//	============

void MainWidgetView::konsoleClose( )
{	if( m_konsoleWidget == NULL )
	{	Q_ASSERT( false );
		return;
	}

	removePage( m_konsoleWidget );
	m_hasKonsole    = false;
	m_konsoleWidget = NULL;

	emit( hasKonsole( false ));
}

//	konsoleOpen
//	===========

void MainWidgetView::konsoleOpen( )
{	if( m_konsoleWidget != NULL )
	{	Q_ASSERT( false );
		return;
	}

	m_hasKonsole = true;
	createKonsolePart( );

	if( m_konsoleWidget == NULL )
	{	m_hasKonsole = false;
		return;
	}

	m_konsoleWidget->show( );
	insertTab( m_konsoleWidget, "Konsole" );
	emit( hasKonsole( true ));
}

//	polish
//	======

void MainWidgetView::polish( )
{	KTabWidget::polish( );

	KConfig* config = kapp->config( );
	config->setGroup( "General_" + m_client->clientName );
	setCurrentPage( config->readNumEntry( "currentPageIndex", 0 ));
}

//	saveSettings
//	============

void MainWidgetView::saveSettings( const bool clientSpecific )
{	m_messagesView->saveSettings( clientSpecific );
	m_projectsView->saveSettings( clientSpecific );
	m_resultsView->saveSettings( clientSpecific );
	m_statisticsView->saveSettings( clientSpecific );
	m_transfersView->saveSettings( clientSpecific );

	KConfig* config = kapp->config( );
	config->setGroup( "General_" + m_client->clientName );
	config->writeEntry( "currentPageIndex", currentPageIndex( ));
	config->writeEntry( "hasKonsole", m_hasKonsole );
}

//	secondsToString
//	===============

QString MainWidgetView::secondsToString( double seconds )
{	ushort hh = byte( seconds / 3600 );
	seconds  -= 3600 * hh;
	byte mm   = byte( seconds / 60 );
	byte ss   = byte( uint( seconds ) % 60 );

	QString text;
	text.sprintf( "%d:%2.2d:%2.2d", hh, mm, ss );
	return text;
}

//	showEvent
//	=========

void MainWidgetView::showEvent( QShowEvent* event )
{	KTabWidget::showEvent( event );
	emit( hasKonsole( m_konsoleWidget ));
}

//	slotKonsolePartDestroyed
//	========================

void MainWidgetView::slotKonsolePartDestroyed( )
{	removePage( m_konsoleWidget );
	m_hasKonsole = false;
	emit( hasKonsole( false ));
}

//	toolTipText
//	===========

QString MainWidgetView::toolTipText( )
{	return m_resultsView->toolTipText( );
}

//	update
//	======

bool MainWidgetView::update( )
{	if( !m_haveStateData )
	{	if( !getState( m_client ))
			return false;

		m_haveStateData = true;
	}

	if( !m_messagesView->update( ))
		return false;

	if( !m_projectsView->update( ))
		return false;

	if( !m_resultsView->update( ))
		return false;

	if( !m_statisticsView->update( ))
		return false;

	if( !m_transfersView->update( ))
		return false;

	return true;
}

//	updateProjects
//	==============

void MainWidgetView::updateProjects( )
{	m_projectsView->slotUpdateProjects( );
}
