/***************************************************************************
 *                                                                         *
 *   Copyright (C) 2006 by Kevin Gilbert                                   *
 *   kev.gilbert@cdu.edu.au                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#include <qgrid.h>
#include <qlabel.h>
#include <qpushbutton.h>
#include <qstringlist.h>
#include <qtimer.h>
#include <qvbox.h>

#include <kapplication.h>
#include <kconfig.h>
#include <kglobalsettings.h>
#include <klocale.h>

#include "mainwidget.h"
#include "messagescontrolview.h"
#include "messagestandarddlg.h"


//	constructor
//	===========

MessagesControlView::MessagesControlView( const CPtr client, QWidget* parent, const char* name )
: QScrollView( parent, name ),
  m_client( client )
{	createLayout( );
	QTimer::singleShot( 0, this, SLOT( slotFinaliseInitialisation( )));
}

//	activateFilters
//	===============

void MessagesControlView::activateFilters( const QString& activeFilters )
{	QStringList filterList = QStringList::split( "\034", activeFilters );
	QPtrList<MessageFilter> filters;

	for( QStringList::Iterator it = filterList.begin( ); it != filterList.end( ); ++it )
		for( MFCBPtr checkBox = m_filterCheckBoxes.first( ); checkBox != NULL; checkBox = m_filterCheckBoxes.next( ))
			if( *it == checkBox->text( ))
			{	checkBox->setChecked( true );
				filters.append( checkBox->filter( ));
				break;
			}

	emit( activateFilters( filters ));
}

//	addFilter
//	=========

void MessagesControlView::addFilter( const MFPtr filter )
{	if( !filter->global( ) && filter->clientName( ) != m_client->clientName )
		return;

	MFCBPtr checkBox = new MessagesFilterCheckBox( filter, m_filtersGrid, "filter checkbox" );
	m_filterCheckBoxes.append( checkBox );

	connect( checkBox, SIGNAL( filterClicked( )),                   SLOT( slotActivateFilter( )));
	connect( checkBox, SIGNAL( filterCopy( const MFPtr )),          SLOT( slotFilterCopy( const MFPtr )));
	connect( checkBox, SIGNAL( filterDelete( const MFPtr )),        SLOT( slotFilterDelete( const MFPtr )));
	connect( checkBox, SIGNAL( filterDownClicked( const MFCBPtr )), SLOT( slotFilterDownClicked( const MFCBPtr )));
	connect( checkBox, SIGNAL( filterEdit( const MFPtr )),          SLOT( slotFilterEdit( const MFPtr )));
	connect( checkBox, SIGNAL( filterUpClicked( const MFCBPtr )),   SLOT( slotFilterUpClicked( const MFCBPtr )));
}

//	createLayout
//	============

void MessagesControlView::createLayout( )
{	byte spacing = fontMetrics( ).width( "XX" );

	QVBox* vBox = new QVBox( viewport( ));
	vBox->setEraseColor( viewport( )->eraseColor( ));
	vBox->setMargin( spacing / 2 );
	vBox->setSpacing( spacing );

	addChild( vBox );

	m_copyAll      = new QPushButton( i18n( "Copy all"      ), vBox, "copy all button"      );
	m_copySelected = new QPushButton( i18n( "Copy selected" ), vBox, "copy selected button" );

	QWidget* spacer1 = new QWidget( vBox );
	spacer1->setEraseColor( KGlobalSettings::buttonBackground( ));
	vBox->setStretchFactor( spacer1, 10 );

	m_filtersGrid = new QGrid( 3, Qt::Horizontal, vBox, "filters grid" );
	m_filtersGrid->setMargin( spacing / 4 );
	m_filtersGrid->setSpacing( spacing / 4 );

	QWidget* spacer2 = new QWidget( vBox );
	spacer2->setEraseColor( KGlobalSettings::buttonBackground( ));
	vBox->setStretchFactor( spacer2, 10 );

	m_filterButtonAdd = new QPushButton( i18n( "Add filter"  ), vBox, "add filter button"  );

	m_copySelected->setEnabled( false );

	connect( m_copyAll,         SIGNAL( clicked( )), SIGNAL( copyAllClicked( )));
	connect( m_copySelected,    SIGNAL( clicked( )), SIGNAL( copySelectedClicked( )));
	connect( m_filterButtonAdd, SIGNAL( clicked( )), SLOT( slotFilterAdd( )));
}

//	findFilter
//	==========

MFPtr MessagesControlView::findFilter( const QString& filterName )
{	for( MFPtr filter = GetFilters.first( ); filter != NULL; filter = GetFilters.next( ))
		if( filterName == filter->filterName( ))
			return filter;

	return NULL;
}

//	filtersUpdated
//	==============

void MessagesControlView::filtersUpdated( )
{	saveSettings( );
	slotFinaliseInitialisation( );
}

//	getActiveFilters
//	================

QString MessagesControlView::getActiveFilters( )
{	QString activeFilters;

	for( QCheckBox* checkBox = m_filterCheckBoxes.first( ); checkBox != NULL; checkBox = m_filterCheckBoxes.next( ))
		if( checkBox->isChecked( ))
		{	if( !activeFilters.isEmpty( ))
				activeFilters += "\034";

			activeFilters += checkBox->text( ).remove( '&' );
		}

	return activeFilters;
}

//	getAllFilters
//	=============

QString MessagesControlView::getAllFilters( )
{	QString allFilters;

	for( QCheckBox* checkBox = m_filterCheckBoxes.first( ); checkBox != NULL; checkBox = m_filterCheckBoxes.next( ))
	{	if( !allFilters.isEmpty( ))
			allFilters += "\034";

		allFilters += checkBox->text( ).remove( '&' );
	}

	return allFilters;
}

//	initFilterView
//	==============

void MessagesControlView::initFilterView( QPtrList<MessageFilter>& filters )
{	QString activeFilters = getActiveFilters( );

	m_filterCheckBoxes.setAutoDelete( true );
	m_filterCheckBoxes.clear( );
	m_filterCheckBoxes.setAutoDelete( false );

	for( MFPtr filter = filters.first( ); filter != NULL; filter = filters.next( ))
		addFilter( filter );

	if( m_filterCheckBoxes.count( ))
	{	m_filterCheckBoxes.getFirst( )->enableUp( false );
		m_filterCheckBoxes.getLast( )->enableDown( false );
	}

	activateFilters( activeFilters );
}

//	saveSettings
//	============

void MessagesControlView::saveSettings( )
{	KConfig* config = kapp->config( );
	config->setGroup( "MessagesControl_" + m_client->clientName );

	config->writeEntry( "activeFilters", getActiveFilters( ));
	config->writeEntry( "filterOrder",   getAllFilters( ));

	config->sync( );
}

//	slotActivateFilter
//	==================

void MessagesControlView::slotActivateFilter( )
{	saveSettings( );
	activateFilters( getActiveFilters( ));
}

//	slotEnableCopySelectedButton
//	============================

void MessagesControlView::slotEnableCopySelectedButton( const bool b )
{	m_copySelected->setEnabled( b );
}

//	slotFilterAdd
//	=============

void MessagesControlView::slotFilterAdd( )
{	MFPtr newFilter = new MessageFilter( QString::null );
	MessageFilterDlg dlg( newFilter, MessageFilterDlg::TypeAdd, this, "message filter dlg" );

	connect( &dlg,
			 SIGNAL( applyFilter( const MFPtr, const MessageFilterDlg::EditType )),
			 SLOT( slotFilterApply( const MFPtr, const MessageFilterDlg::EditType )));

	m_addOrCopyApplied = false;
	dlg.exec( );

	if( !m_addOrCopyApplied )
		delete newFilter;
}

//	slotFilterApply
//	===============

void MessagesControlView::slotFilterApply( const MFPtr filter, const MessageFilterDlg::EditType editType )
{	if( !filter->global( ))
		filter->clientName( m_client->clientName );

	switch( editType )
	{	case MessageFilterDlg::TypeAdd:
		case MessageFilterDlg::TypeCopy:
			if( !m_addOrCopyApplied )
			{	m_addOrCopyApplied = true;
				GetFilters.append( filter );
				addFilter( filter );
			}

			break;

		case MessageFilterDlg::TypeEdit:
			break;

		default:
			fprintf( stderr, "internal error in MessagesControlView::slotFilterApply - unexpected edit type %d\n", editType );
			abort( );
	}

	MainWidget::self( )->filtersUpdated( );
}

//	slotFilterCopy
//	==============

void MessagesControlView::slotFilterCopy( const MFPtr filter )
{	MFPtr newFilter = new MessageFilter( filter );
	MessageFilterDlg dlg( newFilter, MessageFilterDlg::TypeCopy, this, "message filter dlg" );

	connect( &dlg,
			 SIGNAL( applyFilter( const MFPtr, const MessageFilterDlg::EditType )),
			 SLOT( slotFilterApply( const MFPtr, const MessageFilterDlg::EditType )));

	m_addOrCopyApplied = false;
	dlg.exec( );

	if( !m_addOrCopyApplied )
		delete newFilter;
}

//	slotFilterDelete
//	================

void MessagesControlView::slotFilterDelete( const MFPtr filter )
{	GetFilters.remove( filter );
	MainWidget::self( )->filtersUpdated( );
	saveSettings( );
}

//	slotFilterDownClicked
//	=====================

void MessagesControlView::slotFilterDownClicked( const MFCBPtr checkBox )
{	byte i = m_filterCheckBoxes.find( checkBox );
	m_filterCheckBoxes.remove( );
	m_filterCheckBoxes.insert( i + 1, checkBox );

	QPtrList<MessageFilter> filters;

	for( MFCBPtr checkBox = m_filterCheckBoxes.first( ); checkBox != NULL; checkBox = m_filterCheckBoxes.next( ))
		filters.append( checkBox->filter( ));

	initFilterView( filters );
	saveSettings( );
}

//	slotFilterEdit
//	==============

void MessagesControlView::slotFilterEdit( const MFPtr filter )
{	if( filter->isStandardFilter( ))
	{	MessageStandardDlg dlg( this, "standard filter dlg" );
		dlg.exec( );
	}
	else
	{	MessageFilterDlg dlg( filter, MessageFilterDlg::TypeEdit, this, "message filter dlg" );
		connect( &dlg,
				 SIGNAL( applyFilter( const MFPtr, const MessageFilterDlg::EditType )),
				 SLOT( slotFilterApply( const MFPtr, const MessageFilterDlg::EditType )));
		dlg.exec( );
	}
}

//	slotFilterUpClicked
//	===================

void MessagesControlView::slotFilterUpClicked( const MFCBPtr checkBox )
{	byte i = m_filterCheckBoxes.find( checkBox );
	m_filterCheckBoxes.remove( );
	m_filterCheckBoxes.insert( i - 1, checkBox );

	QPtrList<MessageFilter> filters;

	for( MFCBPtr checkBox = m_filterCheckBoxes.first( ); checkBox != NULL; checkBox = m_filterCheckBoxes.next( ))
		filters.append( checkBox->filter( ));

	initFilterView( filters );
	saveSettings( );
}

//	slotFinaliseInitialisation
//	==========================

void MessagesControlView::slotFinaliseInitialisation( )
{	KConfig* config = kapp->config( );
	config->setGroup( "MessagesControl_" + m_client->clientName );

	QStringList filterList = QStringList::split( "\034", config->readEntry( "filterOrder" ));
	QPtrList<MessageFilter> allFilters = GetFilters;
	QPtrList<MessageFilter> filters;

	for( QStringList::Iterator it = filterList.begin( ); it != filterList.end( ); ++it )
		for( MFPtr filter = allFilters.last( ); filter != NULL; filter = allFilters.prev( ))
			if( *it == filter->filterName( ))
			{	filters.append( filter );
				allFilters.remove( filter );
				break;
			}

	for( MFPtr filter = allFilters.last( ); filter != NULL; filter = allFilters.prev( ))
		filters.append( filter );

	initFilterView( filters.count( ) ? filters :  GetFilters );
	activateFilters( config->readEntry( "activeFilters" ));
}
