/***************************************************************************
 *                                                                         *
 *   Copyright (C) 2006 by Kevin Gilbert                                   *
 *   kev.gilbert@cdu.edu.au                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#include <time.h>

#include <qapplication.h>
#include <qclipboard.h>
#include <qregexp.h>
#include <qstringlist.h>
#include <qtimer.h>

#include <kdebug.h>
#include <kglobalsettings.h>
#include <klocale.h>

#include "gui_rpc_client.h"
#include "mainwidget.h"
#include "messageslistview.h"
#include "messagesviewitem.h"

//	macros
//	======

#define AtBottomStart	bool atBottom = !contentsHeight( ) || (visibleHeight( ) >= (contentsHeight( ) - contentsY( ) - horizontalScrollBar( )->height( )))

#define AtBottomEnd		if( atBottom )	\
							m_timer->start( 0, true )

//	constructor
//	===========

MessagesListView::MessagesListView( const CPtr client, QWidget* parent, const char* name )
: KListView( parent, name ),
  SeqNoInc( 0.001 ),
  m_client( client ),
  m_haveHide( false ),
  m_haveShow( false ),
  m_seqNo( 0 ),
  m_timer( new QTimer( this ))
{	setAllColumnsShowFocus( true );
	setSelectionMode( QListView::Extended );
	setRootIsDecorated( true );
	setShowSortIndicator( true );

	addColumn( i18n( "Project" ));
	addColumn( i18n( "Time"    ));
	addColumn( i18n( "Message" ));

	for( byte i = 0; i < columns( ); i++ )
		setColumnWidthMode( i, QListView::Manual );

	connect( this, SIGNAL( selectionChanged( )), SLOT( slotSelectionChanged( )));
	connect( m_timer, SIGNAL( timeout( )), SLOT( slotScrollToEnd( )));
}

//	destructor
//	==========

MessagesListView::~MessagesListView( )
{	delete m_timer;
}

//	applyFilter
//	===========

bool MessagesListView::applyFilter( const MFPtr filter, const MVIPtr item )
{	if( filter->isStandardFilter( ))
	{	applyStandardFilter( item );
		return true;
	}

	bool match;

	if( filter->isRegexp( ))
		match = item->text( 0 ).contains( QRegExp( filter->filterText( ), filter->caseSensitive( ))) ||
				item->text( 1 ).contains( QRegExp( filter->filterText( ), filter->caseSensitive( ))) ||
				item->text( 2 ).contains( QRegExp( filter->filterText( ), filter->caseSensitive( )));
	else
		match = item->text( 0 ).contains( filter->filterText( ), filter->caseSensitive( )) ||
				item->text( 1 ).contains( filter->filterText( ), filter->caseSensitive( )) ||
				item->text( 2 ).contains( filter->filterText( ), filter->caseSensitive( ));

	if( !match )
		return false;

	switch( filter->action( ))
	{	case FilterActionColour:
			item->colour( filter->colour( ));
			break;

		case FilterActionHide:
			item->setVisible( false );
			break;

		case FilterActionShow:
			item->colour( filter->colour( ));
			item->setVisible( true );
			break;
	}

	return true;
}

//	applyFilters
//	============

void MessagesListView::applyFilters( MVIPtr item )
{	for( MFPtr filter = m_filters.first( ); filter != NULL; filter = m_filters.next( ))
	{	if( applyFilter( filter, item ))
			return;

		if( m_haveHide )
			item->setVisible( true );
		else
		if( m_haveShow )
			item->setVisible( false );
	}
}

//	applyStandardFilter
//	===================

void MessagesListView::applyStandardFilter( const MVIPtr item )
{	StdFilter std = MainWidget::self( )->standardFilter( );

	switch( item->priority( ))
	{	case MSG_PRIORITY_ALERT_ERROR: applyStandardFilter( item, std.alertErrorAction, std.alertErrorColour ); break;
		case MSG_PRIORITY_ALERT_INFO:  applyStandardFilter( item, std.alertInfoAction,  std.alertInfoColour  ); break;
		case MSG_PRIORITY_ERROR:       applyStandardFilter( item, std.errorAction,      std.errorColour      ); break;
		case MSG_PRIORITY_INFO:        applyStandardFilter( item, std.infoAction,       std.infoColour       ); break;
		case MSG_PRIORITY_USER:        applyStandardFilter( item, std.userAction,       std.userColour       ); break;

		default:
			applyStandardFilter( item, std.errorAction, std.errorColour );
			break;
	}
}

//	applyStandardFilter
//	===================

void MessagesListView::applyStandardFilter( const MVIPtr item, const FilterAction action, const QColor colour )
{	switch( action )
	{	case FilterActionColour:
			item->colour( colour );
			break;

		case FilterActionHide:
			item->setVisible( false );
			break;

		case FilterActionShow:
			item->colour( colour );
			item->setVisible( true );
			break;
	}
}

//	authenticationFinished
//	======================

void MessagesListView::authenticationFinished( const CPtr /* client */, const bool successful )
{	AtBottomStart;

	QString text = i18n( "Authentication completed - " + QString( successful ? "successfully" : "unsuccessfully" ));
	m_seqNo += SeqNoInc;
	MVIPtr item = new MessagesViewItem( 1, this, MSG_PRIORITY_USER, "KBoincMgr", time( NULL ), m_seqNo, text );
	applyFilters( item );

	AtBottomEnd;
}

//	authenticationStarted
//	=====================

void MessagesListView::authenticationStarted( const CPtr /* client */)
{	AtBottomStart;

	QString text = i18n( "Authentication started" );
	m_seqNo += SeqNoInc;
	MVIPtr item = new MessagesViewItem( 1, this, MSG_PRIORITY_USER, "KBoincMgr", time( NULL ), m_seqNo, text );
	applyFilters( item );

	AtBottomEnd;
}

//	autoProjectUpdateStarted
//	========================

void MessagesListView::autoProjectUpdateStarted( )
{	AtBottomStart;

	m_seqNo += SeqNoInc;
	MVIPtr item = new MessagesViewItem( 1, this, MSG_PRIORITY_USER, "KBoincMgr", time( NULL ), m_seqNo, i18n( "Automatic Project Update started" ));
	applyFilters( item );

	AtBottomEnd;
}

//	filtersClear
//	============

void MessagesListView::filtersClear( )
{	m_filters.clear( );

	for( QListViewItem* item = firstChild( ); item != NULL; item = item->nextSibling( ))
	{	item->setVisible( true );
		MVIPtr( item )->colour( KGlobalSettings::textColor( ));
	}
}

//	slotActivateFilters
//	===================

 void MessagesListView::slotActivateFilters( QPtrList<MessageFilter>& activeFilters )
{	filtersClear( );

	m_filters  = activeFilters;
	m_haveHide = false;
	m_haveShow = false;

	for( MFPtr filter = m_filters.first( ); filter != NULL; filter = m_filters.next( ))
	{	if( filter->action( ) == FilterActionHide )
			m_haveHide = true;

		if( filter->action( ) == FilterActionShow )
			m_haveShow = true;
	}

	for( QListViewItem* item = firstChild( ); item != NULL; item = item->nextSibling( ))
		applyFilters( MVIPtr( item ));

	updateContents( );
}

//	slotCopyAllClicked
//	==================

void MessagesListView::slotCopyAllClicked( )
{	QString text;

	for( QListViewItem* item = firstChild( ); item != NULL; item = item->nextSibling( ))
		text += item->text( 0 ) + ", " + item->text( 1 ) + ", " + item->text( 2 ) + '\n';

    QApplication::clipboard( )->setText( text, QClipboard::Clipboard );
}

//	slotCopySelectedClicked
//	=======================

void MessagesListView::slotCopySelectedClicked( )
{	QString text;

	for( QListViewItem* item = firstChild( ); item != NULL; item = item->nextSibling( ))
		if( item->isSelected( ))
			text += item->text( 0 ) + ", " + item->text( 1 ) + ", " + item->text( 2 ) + '\n';

    QApplication::clipboard( )->setText( text, QClipboard::Clipboard );
}

//	slotSelectionChanged
//	====================

void MessagesListView::slotSelectionChanged( )
{	for( QListViewItem* item = firstChild( ); item != NULL; item = item->nextSibling( ))
		if( item->isSelected( ))
		{	emit( enableCopySelectedButton( true ));
			return;
		}

	emit( enableCopySelectedButton( false ));
}

//	slotScrollToEnd - this is required 'coz if done before window is drawn with horizontal scroll bar,
//	===============   the height of the scroll bar is not taken into account!

void MessagesListView::slotScrollToEnd( )
{	ensureItemVisible( lastItem( ));
}

//	update
//	======

bool MessagesListView::update( )
{	MESSAGES messages;
	int statr = m_client->rpcClient.get_messages( int( m_seqNo ), messages );

	if( statr )
	{	m_client->reason.sprintf( i18n( "get_messages returned error %d" ), statr );
		return false;
	}

	uint noOfMessages = messages.messages.size( );

	if( !noOfMessages )
		return true;

	bool projectsUpdateDone = false;
	AtBottomStart;

	for( uint i = 0; i < noOfMessages; i++ )
	{	int priority;

		if( QString( messages.messages[ i ]->body.c_str( )).contains( "not requesting", false ))
			priority = messages.messages[ i ]->priority;
		else
		if( QString( messages.messages[ i ]->body.c_str( )).contains( "reporting", false ))
				priority = MSG_PRIORITY_ALERT_INFO;
		else
		if( QString( messages.messages[ i ]->body.c_str( )).contains( "Computation", true ) &&
			QString( messages.messages[ i ]->body.c_str( )).contains( "finished", true ))
				priority = MSG_PRIORITY_ALERT_INFO;
		else
		{	if( !projectsUpdateDone && m_seqNo != 0 &&
				QString( messages.messages[ i ]->body.c_str( )).contains( "Scheduler request failed: Error 401", true ))
				{	projectsUpdateDone = true;
					emit( doProjectsUpdate( ));
				}

			priority = messages.messages[ i ]->priority;
		}

		QStringList list = QStringList::split( '\n', messages.messages[ i ]->body.c_str( ));
		MVIPtr item = new MessagesViewItem( 1,
											this,
											priority,
											messages.messages[ i ]->project.c_str( ),
											messages.messages[ i ]->timestamp,
											messages.messages[ i ]->seqno,
											list[ 0 ]);
		applyFilters( item );
		MVIPtr sibling = NULL;

		for( uint j = 1; j < list.count( ); j++ )
		{	m_seqNo += SeqNoInc;

			if( sibling == NULL )
				sibling = new MessagesViewItem( item, priority, m_seqNo, list[ j ]);
			else
				sibling = new MessagesViewItem( item, sibling, priority, m_seqNo, list[ j ]);

			applyFilters( sibling );
		}
	}

	AtBottomEnd;

	m_seqNo = messages.messages[ noOfMessages - 1 ]->seqno;
	return true;
}
