/***************************************************************************
 *                                                                         *
 *   Copyright (C) 2006 by Kevin Gilbert                                   *
 *   kev.gilbert@cdu.edu.au                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#include <qptrlist.h>

#include <kdebug.h>
#include <klocale.h>

#include "authenticate.h"
#include "projectslistview.h"
#include "projectsviewitem.h"

//	constructor
//	===========

ProjectsListView::ProjectsListView( const CPtr client, QWidget* parent, const char* name )
: KListView( parent, name ),
  m_client( client ),
  m_firstProject( true )
{	setAllColumnsShowFocus( true );
	setSelectionMode( QListView::Single );
	setShowSortIndicator( true );

	addColumn( "Project"           );
	addColumn( "User name"         );
	addColumn( "Team"              );
	addColumn( "Total user credit" );
	addColumn( "Total host credit" );
	addColumn( "Avg. user credit"  );
	addColumn( "Avg. host credit"  );
	addColumn( "Resource share"    );
	addColumn( "Status"            );

	setColumnAlignment( 3, Qt::AlignRight );
	setColumnAlignment( 4, Qt::AlignRight );
	setColumnAlignment( 5, Qt::AlignRight );
	setColumnAlignment( 6, Qt::AlignRight );
	setColumnAlignment( 7, Qt::AlignRight );

	for( byte i = 0; i < columns( ); i++ )
		setColumnWidthMode( i, QListView::Manual );

	connect( this, SIGNAL( selectionChanged( )), SLOT( slotSelectionChanged( )));
}

//	doProjectOp
//	===========

void ProjectsListView::doProjectOp( const PVIPtr item, const Operation op )
{	PROJECTS projects;
	int statr = m_client->rpcClient.get_project_status( projects );

	if( statr )
	{	kdDebug( ) << "ProjectsListView::doProjectOp - get_project_status returned error " << statr << endl;
		return;
	}

	uint i;

	for( i = 0; i < projects.projects.size( ); i++ )
		if( item->projectName( ) == projects.projects[ i ]->project_name.c_str( ))
			break;

	if( i >= projects.projects.size( ))
	{	kdDebug( ) << "ProjectsListView::doProjectOp - project >" << item->projectName( ) << "< no longer current" << endl;
		return;
	}

	QString textOp;

	switch( op )
	{	case Detach:  textOp = "detach";  break;
		case Resume:  textOp = "resume";  break;
		case Reset:   textOp = "reset";   break;
		case Suspend: textOp = "suspend"; break;
		case Update:  textOp = "update";  break;

		default:
			Q_ASSERT( false );
			return;
	}

	statr = m_client->rpcClient.project_op( *projects.projects[ i ], textOp );

	if( statr )
	{	kdDebug( ) << "ProjectsListView::doProjectOp - project_op returned error " << statr << endl;
		return;
	}

	emit( enableButtons( false, false ));
	clearSelection( );
}

//	doProjectUpdates
//	================

void ProjectsListView::doProjectUpdates( )
{	emit( autoProjectUpdateStarted( ));

	for( PVIPtr item = PVIPtr( firstChild( )); item != NULL; item = PVIPtr( item->nextSibling( )))
		doProjectOp( item, Update );
}

//	processNewProjects
//	==================

void ProjectsListView::processNewProjects( QPtrList<PROJECT> newProjectList, const double totalResourceShare )
{	for( PROJECT* project = newProjectList.first( ); project != NULL; project = newProjectList.next( ))
	{	m_client->projectName = project->project_name.c_str( );
		PVIPtr item = new ProjectsViewItem( m_client->projectName, this );
		update( item, project, totalResourceShare );
		emit( guiURLs( project ));
	}

	if( m_firstProject )
	{	m_firstProject = false;
		setSelected( firstChild( ), true );
	}
}

//	slotAuthenticationFinished
//	==========================

void ProjectsListView::slotAuthenticationFinished( const CPtr /* client */, const bool successful )
{	disconnect( m_client->authenticate,
				SIGNAL( authenticationFinished( const CPtr, const bool )),
				this,
				SLOT( slotAuthenticationFinished( const CPtr, const bool )));

	if( successful )
		doProjectUpdates( );
}

//	slotDetachClicked
//	=================

void ProjectsListView::slotDetachClicked( )
{	for( PVIPtr item = PVIPtr( firstChild( )); item != NULL; item = PVIPtr( item->nextSibling( )))
		if( item->isSelected( ))
			doProjectOp( item, Detach );
}

//	slotResetClicked
//	================

void ProjectsListView::slotResetClicked( )
{	for( PVIPtr item = PVIPtr( firstChild( )); item != NULL; item = PVIPtr( item->nextSibling( )))
		if( item->isSelected( ))
			doProjectOp( item, Reset );
}

//	slotSelectionChanged
//	====================

void ProjectsListView::slotSelectionChanged( )
{	for( PVIPtr item = PVIPtr( firstChild( )); item != NULL; item = PVIPtr( item->nextSibling( )))
		if( item->isSelected( ))
		{	emit( currentProject( item->projectName( )));
			emit( enableButtons( true, item->suspended( )));
			return;
		}

	emit( enableButtons( false, false ));
}

//	slotSuspendResume
//	=================

void ProjectsListView::slotSuspendResume( const bool suspended )
{	for( PVIPtr item = PVIPtr( firstChild( )); item != NULL; item = PVIPtr( item->nextSibling( )))
		if( item->isSelected( ))
			doProjectOp( item, suspended ? Resume : Suspend  );
}

//	slotUpdateClicked
//	=================

void ProjectsListView::slotUpdateClicked( )
{	for( PVIPtr item = PVIPtr( firstChild( )); item != NULL; item = PVIPtr( item->nextSibling( )))
		if( item->isSelected( ))
			doProjectOp( item, Update );
}

//	update
//	======

bool ProjectsListView::update( )
{	PROJECTS projects;
	int statr = m_client->rpcClient.get_project_status( projects );

	if( statr )
	{	m_client->reason.sprintf( i18n( "get_project_status returned error %d" ), statr );
		return false;
	}

	QPtrList<PROJECT> newProjectList;
	double totalResourceShare = 0;

	for( uint i = 0; i < projects.projects.size( ); i++ )
		totalResourceShare += projects.projects[ i ]->resource_share;

	for( uint i = 0; i < projects.projects.size( ); i++ )
	{	PVIPtr item;

		for( item = PVIPtr( firstChild( )); item != NULL; item = PVIPtr( item->nextSibling( )))
			if( item->projectName( ) == projects.projects[ i ]->project_name.c_str( ))
				break;

		if( item == NULL )
			newProjectList.append( projects.projects[ i ]);
		else
			update( item, projects.projects[ i ], totalResourceShare );
	}

	if( newProjectList.count( ))
		processNewProjects( newProjectList, totalResourceShare );

	return true;
}

//	update
//	======

void ProjectsListView::update( PVIPtr item, PROJECT* project, const double totalResourceShare )
{	item->setText( 0, project->project_name.c_str( ));
	item->setText( 1, project->user_name.c_str( ));
	item->setText( 2, project->team_name.c_str( ));

	QString text;

	text.sprintf( "%.2f", project->user_total_credit  ); item->setText( 3, text );
	text.sprintf( "%.2f", project->host_total_credit  ); item->setText( 4, text );
	text.sprintf( "%.2f", project->user_expavg_credit ); item->setText( 5, text );
	text.sprintf( "%.2f", project->host_expavg_credit ); item->setText( 6, text );

	text.sprintf( "%.1f (%.1f%%)", project->resource_share, 100 * project->resource_share / totalResourceShare );
	item->setText( 7, text );

	if( project->dont_request_more_work )
		item->setText( 8, "not requesting new work" );
	else
	if( project->master_url_fetch_pending )
		item->setText( 8, "matser file fetch pending" );
	else
	if( project->sched_rpc_pending )
		item->setText( 8, "rpc scheduling pending" );
	else
	if( project->scheduler_rpc_in_progress )
		item->setText( 8, "rpc scheduling in progress" );
	else
	if( project->suspended_via_gui )
	{	item->setText( 8, "suspended by gui" );
		item->suspended( true );

		if( item == currentItem( ))
			emit( projectSuspended( true ));
	}
	else
	{	item->setText( 8, "normal" );
		item->suspended( false );

		if( item == currentItem( ))
			emit( projectSuspended( false ));
	}
}

//	updateProjects
//	==============

void ProjectsListView::updateProjects( )
{	if( !m_client->authenticate->autoAuthenticate( ))
	{	doProjectUpdates( );
		return;
	}

	connect( m_client->authenticate,
			 SIGNAL( authenticationFinished( const CPtr, const bool )),
			 SLOT( slotAuthenticationFinished( const CPtr, const bool )));

	m_client->authenticate->authenticate( );
}
