/***************************************************************************
 *                                                                         *
 *   Copyright (C) 2006 by Kevin Gilbert                                   *
 *   kev.gilbert@cdu.edu.au                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#include <qdatetime.h>

#include <kdebug.h>
#include <kglobal.h>
#include <klocale.h>

#include "gui_rpc_client.h"
#include "mainwidgetview.h"
#include "result_state.h"
#include "resultslistview.h"

//	constructor
//	===========

ResultsListView::ResultsListView( const CPtr client, QWidget* parent, const char* name )
: KListView( parent, name ),
  m_client( client )
{	setAllColumnsShowFocus( true );
	setSelectionMode( QListView::Extended );
	setShowSortIndicator( true );

	addColumn( "Project"         );
	addColumn( "Application"     );
	addColumn( "Name"            );
	addColumn( "CPU time"        );
	addColumn( "Progress"        );
	addColumn( "To complete"     );
	addColumn( "Report deadline" );
	addColumn( "Status"          );

	setColumnAlignment( 3, Qt::AlignRight );
	setColumnAlignment( 4, Qt::AlignRight );
	setColumnAlignment( 5, Qt::AlignRight );

	for( byte i = 0; i < columns( ); i++ )
		setColumnWidthMode( i, QListView::Manual );

	connect( this, SIGNAL( selectionChanged( )), SLOT( slotSelectionChanged( )));
}

//	abortResumeOrSuspendResult
//	==========================

void ResultsListView::abortResumeOrSuspendResult( const Operation op )
{	ResultsViewItem* item;

	for( item = m_results.first( ); item != NULL; item = m_results.next( ))
	{	if( !item->isSelected( ))
			continue;

		switch( op )
		{	case Abort:
				abortResumeOrSuspendResult( Abort, item );
				break;

			case Resume:
				if( item->suspended( ))
					abortResumeOrSuspendResult( Resume, item );

				break;

			case Suspend:
				if( item->running( ))
					abortResumeOrSuspendResult( Suspend, item );
		}
	}

	emit( enableButtons( false, false, false, false ));
	clearSelection( );
}

//	abortResumeOrSuspendResult
//	==========================

void ResultsListView::abortResumeOrSuspendResult( const Operation op, ResultsViewItem* item )
{	RESULTS results;
	int statr = m_client->rpcClient.get_results( results );

	if( statr )
	{	kdDebug( ) << "ResultsListView::abortResumeOrSuspendResult - get_results returned error " << statr << endl;
		return;
	}

	uint i;

	for( i = 0; i < results.results.size( ); i++ )
		if( results.results[ i ]->name.c_str( ) == item->name( ))
			break;

	if( i >= results.results.size( ))
	{	kdDebug( ) << "ResultsListView::abortResumeOrSuspendResult - result >" << item->name( ) << "< no longer current" << endl;
		return;
	}

	QString textOp;

	switch( op )
	{	case Abort:   textOp = "abort";   break;
		case Resume:  textOp = "resume";  break;
		case Suspend: textOp = "suspend"; break;

		default:
			Q_ASSERT( false );
			return;
	}

	statr = m_client->rpcClient.result_op( *results.results[ i ], textOp );

	if( statr )
	{	kdDebug( ) << "ResultsListView::abortResumeOrSuspendResult - result_op returned error " << statr << endl;
		return;
	}

	item->preempted( false );
	item->running( false );
	item->suspended( false );
}

//	processNewResults
//	=================

void ResultsListView::processNewResults( QPtrList<RESULT>& newResultsList )
{	RESULT* result;

	for( result = newResultsList.first( ); result != NULL; result = newResultsList.first( ))
	{	newResultsList.removeFirst( );
		ResultsViewItem* item = new ResultsViewItem( 4, this, result );
		m_results.append( item );
		update( item, result );
	}

	m_results.sort( );
}

//	showOrHideGraphics
//	==================

void ResultsListView::showOrHideGraphics( const QString& projectURL, const QString& resultName, const int graphicsMode )
{	DISPLAY_INFO displayInfo;
	int statr = m_client->rpcClient.show_graphics( projectURL, resultName, graphicsMode, displayInfo );

	if( statr )
		kdDebug( ) << "ResultsListView::showOrHideGraphics - show_graphics returned error " << statr << endl;
}

//	slotAbortClicked
//	================

void ResultsListView::slotAbortClicked( )
{	abortResumeOrSuspendResult( Abort );
}

//	slotGraphicsHideClicked
//	=======================

void ResultsListView::slotGraphicsHideClicked( )
{	for( ResultsViewItem* item = m_results.first( ); item != NULL; item = m_results.next( ))
		if( item->isSelected( ))
			showOrHideGraphics( item->projectURL( ), item->name( ), MODE_HIDE_GRAPHICS );

	emit( enableButtons( false, false, false, false ));
	clearSelection( );
}

//	slotGraphicsShowClicked
//	=======================

void ResultsListView::slotGraphicsShowClicked( )
{	for( ResultsViewItem* item = m_results.first( ); item != NULL; item = m_results.next( ))
		if( item->isSelected( ))
			showOrHideGraphics( item->projectURL( ), item->name( ), MODE_WINDOW );

	emit( enableButtons( false, false, false, false ));
	clearSelection( );
}

//	slotSelectionChanged
//	====================

void ResultsListView::slotSelectionChanged( )
{	bool abort    = false;;
	bool graphics = false;;
	bool resume   = false;;
	bool suspend  = false;

	for( ResultsViewItem* item = m_results.first( ); item != NULL; item = m_results.next( ))
		if( item->isSelected( ))
		{	if( item->preempted( ))
				abort = true;
			else
			if( item->running( ))
			{	abort    = true;
				graphics = true;
				suspend  = true;
			}
			else
			if( item->suspended( ))
			{	abort  = true;
				resume = true;
			}
			else
				abort = true;
		}

	emit( enableButtons( abort, graphics, resume, suspend ));
}

//	slotResumeClicked
//	==================

void ResultsListView::slotResumeClicked( )
{	abortResumeOrSuspendResult( Resume );
}

//	slotSuspendClicked
//	==================

void ResultsListView::slotSuspendClicked( )
{	abortResumeOrSuspendResult( Suspend );
}

//	toolTipText
//	===========

QString ResultsListView::toolTipText( )
{	uint    earliestPreempted = uint( -1 );
	uint    earliestRunning   = uint( -1 );
	ushort  finished          = 0;
	ushort  other             = 0;
	ushort  preempted         = 0;
	ushort  running           = 0;
	ushort  suspended         = 0;
	ushort  uploading         = 0;
	ushort  waiting           = 0;

	for( ResultsViewItem* item = m_results.first( ); item != NULL; item = m_results.next( ))
	{	if( item->finished( ))
			finished++;
		else
		if( item->preempted( ))
		{	preempted++;
			earliestPreempted = _min_( earliestPreempted, item->remainingCPU( ));
		}
		else
		if( item->running( ))
		{	running++;
			earliestRunning = _min_( earliestRunning, item->remainingCPU( ));
		}
		else
		if( item->suspended( ))
			suspended++;
		else
		if( item->uploading( ))
			uploading++;
		else
		if( item->waiting( ))
			waiting++;
		else
			other++;
	}

	bool first = true;
	QString text;

	if( finished ) 
	{	if( first ) first = false; else text += ", ";
		text += QString( "finished %1" ).arg( QString::number( finished ));
	}

	if( uploading ) 
	{	if( first ) first = false; else text += ", ";
		text += QString( "uploading %1" ).arg( QString::number( uploading ));
	}

	if( running )
	{	if( first ) first = false; else text += ", ";
		text += QString( "running %1 (earliest %2)" ).arg( QString::number( running )).arg( MainWidgetView::secondsToString( earliestRunning ));
	}

	if( preempted )
	{	if( first ) first = false; else text += ", ";
		text += QString( "preempted %1 (earliest %2)" ).arg( QString::number( preempted )).arg( MainWidgetView::secondsToString( earliestPreempted ));
	}

	if( suspended )
	{	if( first ) first = false; else text += ", ";
		text += QString( "suspended %1" ).arg( QString::number( suspended ));
	}

	if( waiting )
	{	if( first ) first = false; else text += ", ";
		text += QString( "waiting %1" ).arg( QString::number( waiting ));
	}

	if( other )
	{	if( first ) first = false; else text += ", ";
		text += QString( "other %1" ).arg( QString::number( other ));
	}

	return text;
}

//	update
//	======

bool ResultsListView::update( )
{	RESULTS results;
	int statr = m_client->rpcClient.get_results( results );

	if( statr )
	{	m_client->reason.sprintf( i18n( "get_results returned error %d" ), statr );
		return false;
	}

	for( ResultsViewItem* item = m_results.first( ); item != NULL; item = m_results.next( ))
		item->updated( false );

	QPtrList<RESULT> newResultsList;

	for( uint i = 0; i < results.results.size( ); i++ )
	{	ResultsViewItem* item;

		for( item = m_results.first( ); item != NULL; item = m_results.next( ))
		{	statr = QString::compare( item->name( ), results.results[ i ]->name.c_str( ));

			if( !statr )
				break;

			if( statr > 0 )
			{	item = NULL;
				break;
			}
		}

		if( item == NULL )
			newResultsList.append( results.results[ i ]);
		else
		{	update( item, results.results[ i ]);
			item->updated( true );
		}
	}

	for( ResultsViewItem* item = m_results.last( ); item != NULL; item = m_results.prev( ))
		if( !item->updated( ))
		{	m_results.remove( );
			delete item;
		}

	if( newResultsList.count( ))
		processNewResults( newResultsList );

	sort( );
	return true;
}

//	update
//	======

void ResultsListView::update( ResultsViewItem* item, RESULT* result )
{	RESULT* stateResult = m_client->state.lookup_result( result->project_url, result->name );

	if( stateResult == NULL && MainWidgetView::getState( m_client ))
		stateResult = m_client->state.lookup_result( result->project_url, result->name );

	item->setProjectURL( result->project_url );
	item->setProject( 0, stateResult == NULL ? "**** unknown ****" : stateResult->project->project_name.c_str( ));
	item->setRemainingCPU( 5, uint( result->estimated_cpu_time_remaining ));
	item->setText( 1, stateResult == NULL ? "**** unknown ****" : stateResult->app->name.c_str( ));
	item->setText( 2, result->name.c_str( ));

	QDateTime reportDeadline;
	reportDeadline.setTime_t( result->report_deadline );
	item->setText( 6, KGlobal::locale( )->formatDateTime( reportDeadline, true, true ));

	item->finished( false );
	item->preempted( false );
	item->running( false );
	item->suspended( false );
	item->uploading( false );
	item->waiting( false );

	QString cpuTime;
	double  progress;

	if( result->project_suspended_via_gui )
	{	item->setText( 7, "Project suspended by user" );
		cpuTime  = MainWidgetView::secondsToString( result->final_cpu_time );
		progress = 100 * result->fraction_done;
	}
	else
	if( result->ready_to_report )
	{	item->finished( true );
		item->setText( 7, "Ready to report" );
		progress = 100;
		cpuTime  = MainWidgetView::secondsToString( result->final_cpu_time );
	}
	else
	if( result->suspended_via_gui )
	{	item->suspended( true );
		item->setText( 7, "Suspended by user" );
		cpuTime  = MainWidgetView::secondsToString( result->final_cpu_time );
		progress = _max_( 0.01, 100 * result->fraction_done );
	}
	else
	{	QString text;

		switch( result->state )
		{	case RESULT_COMPUTE_ERROR:
				cpuTime  = MainWidgetView::secondsToString( result->final_cpu_time );
				progress = 100;
				text     = "Compute error";
				break;

			case RESULT_FILES_DOWNLOADING:
				progress = 0;
				text     = "Downloading";
				break;

			case RESULT_FILES_UPLOADED:
				cpuTime  = MainWidgetView::secondsToString( result->final_cpu_time );
				progress = 100;
				text     = "Uploaded";
				break;

			case RESULT_FILES_UPLOADING:
				item->uploading( true );
				cpuTime  = MainWidgetView::secondsToString( result->final_cpu_time );
				progress = 100;
				text     = "Uploading";
				break;

			case RESULT_NEW:
				progress = 0;
				text     = "New";
				break;

			case RESULT_FILES_DOWNLOADED:
				switch( result->scheduler_state )
				{	case CPU_SCHED_PREEMPTED:
						item->preempted( true );
						progress = _max_( 0.01, 100 * result->fraction_done );
						text     = "Preempted";
						break;

					case CPU_SCHED_SCHEDULED:
						item->running( true );
						progress = 100 * result->fraction_done;
						text     = "Running";
						break;

					case CPU_SCHED_UNINITIALIZED:
						item->waiting( true );
						progress = 0;
						text     = "Ready to run";
						break;

					default:
						progress = 100 * result->fraction_done;
						text     = "Unknown";
				}

				break;

			default:
				progress = 100 * result->fraction_done;
				text     = "Unknown";
		}

		item->setText( 7, text );
	}

	item->setProgress( progress );

	if( result->active_task )
		item->setText( 3, MainWidgetView::secondsToString( result->current_cpu_time ));
	else
		item->setText( 3, cpuTime  );
}
