/***************************************************************************
 *                                                                         *
 *   Copyright (C) 2006 by Kevin Gilbert                                   *
 *   kev.gilbert@cdu.edu.au                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#include <qbuttongroup.h>
#include <qradiobutton.h>
#include <qvbox.h>

#include <kapplication.h>
#include <kcombobox.h>
#include <kconfig.h>
#include <kdebug.h>
#include <klocale.h>

#include "gui_rpc_client.h"
#include "statisticscontroldata.h"
#include "statisticscontrolview.h"

//	constructor
//	===========

StatisticsControlView::StatisticsControlView( const CPtr client, QWidget* parent, const char* name )
: QScrollView( parent, name ),
  m_client( client )
{	KConfig* config = kapp->config( );
	config->setGroup( "StatisticsGraphControl_" + m_client->clientName );

	QString currentProject = config->readEntry( "currentProject", QString::null );
	ushort  noOfProjects   = config->readNumEntry( "noOfProjects", 0 );
	QString text;

    for ( ushort i = 0; i < noOfProjects; i++ )
	{	QString key = QString( "projectData_%1" ).arg( i );
		m_projectsData.append( new StatisticsControlData( config->readEntry( key, QString::null )));
	}

	createLayout( currentProject );
	finaliseInitialisation( );
}

//	compareProjects
//	===============

int StatisticsControlView::compareProjects( const void* project1, const void* project2 )
{	return (*SCDPPtr( project1 ))->compare( *SCDPPtr( project2 ));
}

//	createLayout
//	============

void StatisticsControlView::createLayout( const QString& currentProject )
{	byte spacing = fontMetrics( ).width( "XX" );

	QVBox* vBox = new QVBox( viewport( ));
	vBox->setEraseColor( viewport( )->eraseColor( ));
	vBox->setMargin( spacing / 2 );
	vBox->setSpacing( spacing );

	addChild( vBox );
	PROJECTS projects;
	int statr = m_client->rpcClient.get_project_status( projects );

	if( statr )
	{	m_client->reason.sprintf( i18n( "get_project_status returned error %d" ), statr );
		return;
	}

	ushort size = projects.projects.size( );

	if( !size )
		return;

	QWidget* spacer1 = new QWidget( vBox );
	spacer1->setEraseColor( KGlobalSettings::buttonBackground( ));

	m_projectsComboBox  = new KComboBox( false, vBox, "projects combo box" );
	connect( m_projectsComboBox, SIGNAL( activated( const QString& )), SLOT( slotProjectActivated( const QString& )));

	QWidget* spacer2 = new QWidget( vBox );
	spacer2->setEraseColor( KGlobalSettings::buttonBackground( ));
	vBox->setStretchFactor( spacer2, 10 );

	ushort currentIndex = 0;

	for( ushort i = 0; i <= size; i++ )
	{	QString projectName  = (i == size) ? StatisticsControlData::StatsStackedLit : QString( projects.projects[ i ]->project_name );
		SCDPtr projectData;

    	for( projectData = m_projectsData.first( ); projectData != NULL; projectData = m_projectsData.next( ))
			if( projectName == projectData->projectName( ))
				break;

		if( projectData == NULL )
		{	projectData = new StatisticsControlData( UserTotal, projectName );
			m_projectsData.append( projectData );
		}

		m_projectsComboBox->insertItem( projectName );
		projectData->guiRPCIndex( i );
		projectData->isKnown( true );

		if( i == size )
			projectData->stacked( true );

		if( projectName == currentProject )
			currentIndex = i;
	}

	m_projectsComboBox->setCurrentItem( currentIndex );

	QButtonGroup* group = new QButtonGroup( 4, Qt::Vertical, i18n( "Actions" ), vBox, "button group" );
	m_hostAvg   = new QRadioButton( i18n( "Host Average" ), group, "host average button" );
	m_hostTotal = new QRadioButton( i18n( "Host Total"   ), group, "host total button"   );
	m_userAvg   = new QRadioButton( i18n( "User Average" ), group, "user average button" );
	m_userTotal = new QRadioButton( i18n( "User Total"   ), group, "user total button"   );

	group->setEraseColor( vBox->eraseColor( ));
	m_hostAvg->setEraseColor( vBox->eraseColor( ));
	m_hostTotal->setEraseColor( vBox->eraseColor( ));
	m_userAvg->setEraseColor( vBox->eraseColor( ));
	m_userTotal->setEraseColor( vBox->eraseColor( ));

	switch( findCurrentProject( )->graphType( ))
	{	case HostAvg:   m_hostAvg->setChecked( true );   break;
		case HostTotal: m_hostTotal->setChecked( true ); break;
		case UserAvg:   m_userAvg->setChecked( true );   break;
		case UserTotal: m_userTotal->setChecked( true ); break;
	}

	setTabOrder( m_projectsComboBox, m_hostAvg   );
	setTabOrder( m_hostAvg,          m_hostTotal );
	setTabOrder( m_hostTotal,        m_userAvg   );
	setTabOrder( m_userAvg,          m_userTotal );

	connect( m_hostAvg,   SIGNAL( clicked( )), SLOT( slotHostAvgClicked( )));
	connect( m_hostTotal, SIGNAL( clicked( )), SLOT( slotHostTotalClicked( )));
	connect( m_userAvg,   SIGNAL( clicked( )), SLOT( slotUserAvgClicked( )));
	connect( m_userTotal, SIGNAL( clicked( )), SLOT( slotUserTotalClicked( )));
}

//	finaliseInitialisation
//	======================

void StatisticsControlView::finaliseInitialisation( )
{	ushort  noOfProjects = m_projectsData.count( );
	SCDPPtr sortData     = new SCDPtr[ noOfProjects ];
	ushort  i;
	SCDPtr  projectData;

    for( i = 0, projectData = m_projectsData.first( ); projectData != NULL; i++, projectData = m_projectsData.next( ))
		sortData[ i ] = projectData;

	::qsort( sortData, noOfProjects, sizeof( SCDPtr ), compareProjects );
	m_projectsData.clear( );

	for( i = 0; i < noOfProjects; i++ )
		if( sortData[ i ]->isKnown( ))
			m_projectsData.append( sortData[ i ]);
		else
			delete sortData[ i ];

	delete [] sortData;
	m_projectsData.setAutoDelete( true );
}

//	findCurrentProject
//	==================

SCDPtr StatisticsControlView::findCurrentProject( )
{	QString projectName = m_projectsComboBox->currentText( );
	SCDPtr  projectData;

    for( projectData = m_projectsData.first( ); projectData != NULL; projectData = m_projectsData.next( ))
		if( projectData->projectName( ) == projectName )
			return projectData;

	fprintf( stderr, "internal error in StatisticsControlView::findCurrentProject - cannot find project \"%s\"\n", projectName.ascii( ));
	abort( );
	return NULL;
}

//	saveSettings
//	============

void StatisticsControlView::saveSettings( )
{	KConfig* config = kapp->config( );
	config->setGroup( "StatisticsGraphControl_" + m_client->clientName );
	config->writeEntry( "currentProject", findCurrentProject( )->projectName( ));
	config->writeEntry( "noOfProjects",   m_projectsData.count( ));

	ushort  i;
	SCDPtr  projectData;
	QString text;

    for( i = 0, projectData = m_projectsData.first( ); projectData != NULL; i++, projectData = m_projectsData.next( ))
		if( projectData->isKnown( ))
			config->writeEntry( QString( "projectData_%1" ).arg( i ), projectData->toString( ));
}

//	slotHostAvgClicked
//	==================

void StatisticsControlView::slotHostAvgClicked( )
{	SCDPtr projectData = findCurrentProject( );
	projectData->graphType( HostAvg );
	emit( hostAverageActivated( projectData->projectName( )));
}

//	slotHostTotalClicked
//	====================

void StatisticsControlView::slotHostTotalClicked( )
{	SCDPtr projectData = findCurrentProject( );
	projectData->graphType( HostTotal );
	emit( hostTotalActivated( projectData->projectName( )));
}

//	slotProjectActivated
//	====================

void StatisticsControlView::slotProjectActivated( const QString& /* projectName */)
{	switch( findCurrentProject( )->graphType( ))
	{	case HostAvg:   m_hostAvg->setChecked( true );   slotHostAvgClicked( );   break;
		case HostTotal: m_hostTotal->setChecked( true ); slotHostTotalClicked( ); break;
		case UserAvg:   m_userAvg->setChecked( true );   slotUserAvgClicked( );   break;
		case UserTotal: m_userTotal->setChecked( true ); slotUserTotalClicked( ); break;
	}
}

//	slotUserAvgClicked
//	==================

void StatisticsControlView::slotUserAvgClicked( )
{	SCDPtr projectData = findCurrentProject( );
	projectData->graphType( UserAvg );
	emit( userAverageActivated( projectData->projectName( )));
}

//	slotUserTotalClicked
//	====================

void StatisticsControlView::slotUserTotalClicked( )
{	SCDPtr projectData = findCurrentProject( );
	projectData->graphType( UserTotal );
	emit( userTotalActivated( projectData->projectName( )));
}
