/***************************************************************************
 *                                                                         *
 *   Copyright (C) 2006 by Kevin Gilbert                                   *
 *   kev.gilbert@cdu.edu.au                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#include <math.h>

#include "statisticscontroldata.h"
#include "statisticsdata.h"
#include "statisticsprojectdata.h"

//	constructor
//	===========

StatisticsData::StatisticsData( const PROJECTS& projects,
								QPtrList<StatisticsControlData>* projectsData,
								const QString& clientName,
								const bool stacked,
								QValueList<uint>* removedDataPoints )
: m_projectData( NULL ),
  m_removedDataPoints( removedDataPoints )
{	createDataStructures( projects );
	extractData( projects );
	calcTotals( );
	smoothStacked( stacked );
	calcMinMax( );
	scaleValues( stacked );
	buildToolTipText( projectsData, clientName );
}

//	destructor
//	==========

StatisticsData::~StatisticsData( )
{	deleteProjectData( );
}

//	buildToolTipText
//	================

void StatisticsData::buildToolTipText( QPtrList<StatisticsControlData>* projectsData, const QString& clientName )
{	SCDPtr projectData;

	for( projectData = projectsData->first( ); projectData != NULL; projectData = projectsData->next( ))
		m_projectData[ projectData->guiRPCIndex( )].buildToolTipText( projectData->projectName( ),
																	  clientName,
																	  projectData->stacked( ));
}

//	calcMinMax
//	==========

void StatisticsData::calcMinMax( )
{	for( ushort i = 0; i < m_totalProject; i++ )
		m_projectData[ i ].calcMax( );

	m_projectData[ 0 ].calcMin( );
	SPDPtr stacked = m_projectData + m_noOfProjects;

	stacked->m_minHostAvg   = m_projectData[ 0 ].m_minHostAvg;
	stacked->m_minHostTotal = m_projectData[ 0 ].m_minHostTotal;
	stacked->m_minUserAvg   = m_projectData[ 0 ].m_minUserAvg;
	stacked->m_minUserTotal = m_projectData[ 0 ].m_minUserTotal;

	for( ushort i = 1; i < m_noOfProjects; i++ )
	{	m_projectData[ i ].calcMin( );

		stacked->m_minHostAvg   = _min_( stacked->m_minHostAvg,   m_projectData[ i ].m_minHostAvg   );
		stacked->m_minHostTotal = _min_( stacked->m_minHostTotal, m_projectData[ i ].m_minHostTotal );
		stacked->m_minUserAvg   = _min_( stacked->m_minUserAvg,   m_projectData[ i ].m_minUserAvg   );
		stacked->m_minUserTotal = _min_( stacked->m_minUserTotal, m_projectData[ i ].m_minUserTotal );
	}
}

//	calcTotals
//	==========

void StatisticsData::calcTotals( )
{	m_totalHostAvg   = 0;
	m_totalHostTotal = 0;
	m_totalUserAvg   = 0;
	m_totalUserTotal = 0;

	for( ushort i = 0; i < m_noOfDataPoints; i++ )
	{	m_totalHostAvg   += m_projectData[ m_noOfProjects ].m_projectDateData[ i ].m_hostAvg;
		m_totalHostTotal += m_projectData[ m_noOfProjects ].m_projectDateData[ i ].m_hostTotal;
		m_totalUserAvg   += m_projectData[ m_noOfProjects ].m_projectDateData[ i ].m_userAvg;
		m_totalUserTotal += m_projectData[ m_noOfProjects ].m_projectDateData[ i ].m_userTotal;
	}
}

//	compareDates
//	============

int StatisticsData::compareDates( const void* date1, const void* date2 )
{	if( *uiPtr( date1 ) < *uiPtr( date2 ))
		return -1;

	return *uiPtr( date1 ) != *uiPtr( date2 );
}

//	createDataStructures
//	====================

void StatisticsData::createDataStructures( const PROJECTS& projects )
{	m_noOfProjects = projects.projects.size( );
	m_totalProject = m_noOfProjects + 1;
	ushort totalNoOfDataPoints = 0;

	for( ushort i = 0; i < m_noOfProjects; i++ )
		totalNoOfDataPoints += projects.projects[ i ]->statistics.size( );

	uiPtr  dates = new uint[ totalNoOfDataPoints ];
	ushort k = 0;

	for( ushort i = 0; i < m_noOfProjects; i++ )
	{	PROJECT* project    = projects.projects[ i ];
		ushort   noOfPoints = project->statistics.size( );

		for( uint j = 0; j < noOfPoints; j++ )
			dates[ k++ ] = uint( project->statistics[ j ].day );
	}

	qsort( dates, totalNoOfDataPoints, sizeof( uint ), compareDates );
	uint priorDate   = 0;
	m_noOfDataPoints = 0;

	for( ushort i = 0; i < totalNoOfDataPoints; i++ )
		if( dates[ i ] != priorDate )
		{	m_noOfDataPoints++;
			priorDate = dates[ i ];
		}

	deleteProjectData( );
	m_projectData = new StatisticsProjectData[ m_totalProject ];

	for( ushort i = 0; i < m_totalProject; i++ )
		m_projectData[ i ].initialise( m_noOfDataPoints );

	k = 0;
	priorDate = 0;

	for( ushort i = 0; i < totalNoOfDataPoints; i++ )
		if( dates[ i ] != priorDate )
		{	for( ushort j = 0; j < m_totalProject; j++ )
				m_projectData[ j ].m_projectDateData[ k ].m_date = dates[ i ];

			k++;
			priorDate = dates[ i ];
		}

	delete [] dates;
}

//	date
//	====

uint StatisticsData::date( const ushort projectNo, const ushort i ) const
{	return m_projectData[ projectNo ].m_projectDateData[ i ].m_date;
}

//	deleteProjectData
//	=================

void StatisticsData::deleteProjectData( )
{	if( m_projectData == NULL )
		return;

	delete [] m_projectData;
	m_projectData = NULL;
}

//	extractData
//	===========

void StatisticsData::extractData( const PROJECTS& projects )
{	QValueList<uint>::iterator it;

	for( ushort i = 0; i < m_noOfProjects; i++ )
	{	ushort   k          = 0;
		PROJECT* project    = projects.projects[ i ];
		ushort   noOfPoints = project->statistics.size( );

		for( uint j = 0; j < noOfPoints; j++ )
		{	DAILY_STATS* stats = &project->statistics[ j ];
			uint date = uint( stats->day );

			for( it = m_removedDataPoints->begin( ); it != m_removedDataPoints->end( ); ++it )
				if( (*it) == date )
					break;

			if( it != m_removedDataPoints->end( ))
				continue;

			for( ; date != m_projectData[ i ].m_projectDateData[ k ].m_date; k++ ) ;
			m_projectData[ i ].m_projectDateData[ k ].m_valid = true;

			m_projectData[ i ].m_projectDateData[ k ].m_hostAvg   = stats->host_expavg_credit;
			m_projectData[ i ].m_projectDateData[ k ].m_hostTotal = stats->host_total_credit;
			m_projectData[ i ].m_projectDateData[ k ].m_userAvg   = stats->user_expavg_credit;
			m_projectData[ i ].m_projectDateData[ k ].m_userTotal = stats->user_total_credit;

			m_projectData[ m_noOfProjects ].m_projectDateData[ k ].m_hostAvg   += stats->host_expavg_credit;
			m_projectData[ m_noOfProjects ].m_projectDateData[ k ].m_hostTotal += stats->host_total_credit;
			m_projectData[ m_noOfProjects ].m_projectDateData[ k ].m_userAvg   += stats->user_expavg_credit;
			m_projectData[ m_noOfProjects ].m_projectDateData[ k ].m_userTotal += stats->user_total_credit;

			m_projectData[ i ].m_projectDateData[ k ].m_valid = true;
			m_projectData[ m_noOfProjects ].m_projectDateData[ k ].m_valid = true;
		}
	}
}

//	interpolated
//	============

bool StatisticsData::interpolated( const ushort projectNo, const ushort dateIndex ) const
{	return m_projectData[ projectNo ].m_projectDateData[ dateIndex ].m_interpolated;
}

//	maxDate
//	=======

uint StatisticsData::maxDate( ) const
{	return m_projectData[ 0 ].m_projectDateData[ m_noOfDataPoints - 1 ].m_date;
}

//	minDate
//	=======

uint StatisticsData::minDate( ) const
{	return m_projectData[ 0 ].m_projectDateData[ 0 ].m_date;
}

//	maxValue
//	========

double StatisticsData::maxValue( const ushort projectNo, const GraphType graphType, const bool stacked ) const
{	if( !stacked )
		switch( graphType )
		{	case HostAvg:   return m_projectData[ projectNo ].m_maxHostAvg;
			case HostTotal: return m_projectData[ projectNo ].m_maxHostTotal;
			case UserAvg:   return m_projectData[ projectNo ].m_maxUserAvg;
			case UserTotal: return m_projectData[ projectNo ].m_maxUserTotal;
		}

	double max = 0;	//	just to stop the compiler from whinging

	switch( graphType )
	{	case HostAvg:   max = m_projectData[ 0 ].m_maxHostAvg;   break;
		case HostTotal: max = m_projectData[ 0 ].m_maxHostTotal; break;
		case UserAvg:   max = m_projectData[ 0 ].m_maxUserAvg;   break;
		case UserTotal: max = m_projectData[ 0 ].m_maxUserTotal; break;
	}

	for( ushort i = 1; i < m_totalProject; i++ )
		switch( graphType )
		{	case HostAvg:   max = _max_( max, m_projectData[ i ].m_maxHostAvg   ); break;
			case HostTotal: max = _max_( max, m_projectData[ i ].m_maxHostTotal ); break;
			case UserAvg:   max = _max_( max, m_projectData[ i ].m_maxUserAvg   ); break;
			case UserTotal: max = _max_( max, m_projectData[ i ].m_maxUserTotal ); break;
		}

	return max;
}

//	minValue
//	========

double StatisticsData::minValue( const ushort projectNo, const GraphType graphType, const bool stacked ) const
{	if( !stacked )
		switch( graphType )
		{	case HostAvg:   return m_projectData[ projectNo ].m_minHostAvg;
			case HostTotal: return m_projectData[ projectNo ].m_minHostTotal;
			case UserAvg:   return m_projectData[ projectNo ].m_minUserAvg;
			case UserTotal: return m_projectData[ projectNo ].m_minUserTotal;
		}

	double min = 0;	//	just to stop the compiler from whinging

	switch( graphType )
	{	case HostAvg:   min = m_projectData[ 0 ].m_minHostAvg;   break;
		case HostTotal: min = m_projectData[ 0 ].m_minHostTotal; break;
		case UserAvg:   min = m_projectData[ 0 ].m_minUserAvg;   break;
		case UserTotal: min = m_projectData[ 0 ].m_minUserTotal; break;
	}

	for( ushort i = 0; i < m_totalProject; i++ )
		switch( graphType )
		{	case HostAvg:   min = _min_( min, m_projectData[ i ].m_minHostAvg   ); break;
			case HostTotal: min = _min_( min, m_projectData[ i ].m_minHostTotal ); break;
			case UserAvg:   min = _min_( min, m_projectData[ i ].m_minUserAvg   ); break;
			case UserTotal: min = _min_( min, m_projectData[ i ].m_minUserTotal ); break;
		}

	return min;
}

//	point
//	=====

QRect StatisticsData::point( const ushort projectNo, const ushort dateIndex, const GraphType graphType ) const
{	switch( graphType )
	{	case HostAvg:   return m_projectData[ projectNo ].m_projectDateData[ dateIndex ].m_pointHostAvg;
		case HostTotal: return m_projectData[ projectNo ].m_projectDateData[ dateIndex ].m_pointHostTotal;
		case UserAvg:   return m_projectData[ projectNo ].m_projectDateData[ dateIndex ].m_pointUserAvg;
		case UserTotal: return m_projectData[ projectNo ].m_projectDateData[ dateIndex ].m_pointUserTotal;
	}

	Q_ASSERT( false );
	return QRect( 0, 0, 0, 0 );
}

//	point
//	=====

void StatisticsData::point( const ushort projectNo, const ushort dateIndex, const GraphType graphType, const QRect& rect )
{	switch( graphType )
	{	case HostAvg:   m_projectData[ projectNo ].m_projectDateData[ dateIndex ].m_pointHostAvg   = rect; return;
		case HostTotal: m_projectData[ projectNo ].m_projectDateData[ dateIndex ].m_pointHostTotal = rect; return;
		case UserAvg:   m_projectData[ projectNo ].m_projectDateData[ dateIndex ].m_pointUserAvg   = rect; return;
		case UserTotal: m_projectData[ projectNo ].m_projectDateData[ dateIndex ].m_pointUserTotal = rect; return;
	}
}

//	sameData
//	========

bool StatisticsData::sameData( const PROJECTS& projects ) const
{	if( !m_noOfProjects || m_noOfProjects != projects.projects.size( ))
		return false;

	QValueList<uint>::iterator it;

	double totalHostAvg   = 0;
	double totalHostTotal = 0;
	double totalUserAvg   = 0;
	double totalUserTotal = 0;

	for( ushort i = 0; i < m_noOfProjects; i++ )
	{	PROJECT* project    = projects.projects[ i ];
		ushort   noOfPoints = project->statistics.size( );

		for( uint j = 0; j < noOfPoints; j++ )
		{	DAILY_STATS* stats = &project->statistics[ j ];
			uint date = uint( stats->day );

			for( it = m_removedDataPoints->begin( ); it != m_removedDataPoints->end( ); ++it )
				if( (*it) == date )
					break;

			if( it != m_removedDataPoints->end( ))
				continue;

			totalHostAvg   += stats->host_expavg_credit;
			totalHostTotal += stats->host_total_credit;
			totalUserAvg   += stats->user_expavg_credit;
			totalUserTotal += stats->user_total_credit;
		}
	}

	return _equal_( totalHostAvg,   m_totalHostAvg   ) &&
		   _equal_( totalHostTotal, m_totalHostTotal ) &&
		   _equal_( totalUserAvg,   m_totalUserAvg   ) &&
		   _equal_( totalUserTotal, m_totalUserTotal );
}

//	scaledValue
//	===========

double StatisticsData::scaledValue( const ushort projectNo, const ushort dateIndex, const GraphType graphType ) const
{	switch( graphType )
	{	case HostAvg:   return m_projectData[ projectNo ].m_projectDateData[ dateIndex ].m_scaledHostAvg;
		case HostTotal: return m_projectData[ projectNo ].m_projectDateData[ dateIndex ].m_scaledHostTotal;
		case UserAvg:   return m_projectData[ projectNo ].m_projectDateData[ dateIndex ].m_scaledUserAvg;
		case UserTotal: return m_projectData[ projectNo ].m_projectDateData[ dateIndex ].m_scaledUserTotal;
	}

	Q_ASSERT( false );
	return 0;
}

//	scaleValues
//	===========

void StatisticsData::scaleValues( const bool stacked )
{	for( ushort i = 0; i < m_totalProject; i++ )
		m_projectData[ i ].scaleValues( stacked ? m_projectData[ m_noOfProjects ] : m_projectData[ i ]);
}

//	smoothStacked
//	=============

void StatisticsData::smoothStacked( const bool stacked )
{	if( !stacked )
		return;

	SPDPtr stackedProject = m_projectData + m_noOfProjects;


	for( ushort j = 0; j < m_noOfDataPoints; j++ )
	{	StatisticsProjectData::PDDPtr stackedDateData = stackedProject->m_projectDateData + j;

		for( ushort i = 0; i < m_noOfProjects; i++ )
		{	SPDPtr project = m_projectData + i;

			if( project->m_projectDateData[ j ].m_valid )
				continue;

			ushort next;
			short  prior;

			for( next  = j + 1; next < m_noOfDataPoints && !project->m_projectDateData[ next ].m_valid; next++ ) ;
			for( prior = j - 1; prior >= 0 && !project->m_projectDateData[ prior ].m_valid; prior-- ) ;

			if( prior < 0 && next < m_noOfDataPoints )
			{	stackedDateData->m_hostAvg   += project->m_projectDateData[ next ].m_hostAvg;
				stackedDateData->m_hostTotal += project->m_projectDateData[ next ].m_hostTotal;
				stackedDateData->m_userAvg   += project->m_projectDateData[ next ].m_userAvg;
				stackedDateData->m_userTotal += project->m_projectDateData[ next ].m_userTotal;
			}
			else
			if( prior >= 0 && next >= m_noOfDataPoints - 1 )
			{	stackedDateData->m_hostAvg   += project->m_projectDateData[ prior ].m_hostAvg;
				stackedDateData->m_hostTotal += project->m_projectDateData[ prior ].m_hostTotal;
				stackedDateData->m_userAvg   += project->m_projectDateData[ prior ].m_userAvg;
				stackedDateData->m_userTotal += project->m_projectDateData[ prior ].m_userTotal;
			}
			else
			{	double x  = project->m_projectDateData[ j     ].m_date;
				double x1 = project->m_projectDateData[ prior ].m_date;
				double x2 = project->m_projectDateData[ next  ].m_date;
				double c;
				double m;
				double y1;
				double y2;

				y1 = project->m_projectDateData[ prior ].m_hostAvg;
				y2 = project->m_projectDateData[ next  ].m_hostAvg;
				c  = (x1 * y2 - x2 * y1 ) / (x1 - x2);
				m  = ( y1 - y2) / (x1 - x2);
				stackedDateData->m_hostAvg += m * x + c;

				y1 = project->m_projectDateData[ prior ].m_hostTotal;
				y2 = project->m_projectDateData[ next  ].m_hostTotal;
				c  = (x1 * y2 - x2 * y1 ) / (x1 - x2);
				m  = ( y1 - y2) / (x1 - x2);
				stackedDateData->m_hostTotal += m * x + c;

				y1 = project->m_projectDateData[ prior ].m_userAvg;
				y2 = project->m_projectDateData[ next  ].m_userAvg;
				c  = (x1 * y2 - x2 * y1 ) / (x1 - x2);
				m  = ( y1 - y2) / (x1 - x2);
				stackedDateData->m_userAvg += m * x + c;

				y1 = project->m_projectDateData[ prior ].m_userTotal;
				y2 = project->m_projectDateData[ next  ].m_userTotal;
				c  = (x1 * y2 - x2 * y1 ) / (x1 - x2);
				m  = ( y1 - y2) / (x1 - x2);
				stackedDateData->m_userTotal += m * x + c;
			}

			stackedDateData->m_interpolated = true;
		}
	}
}

//	toolTipText
//	===========

QString StatisticsData::toolTipText( const ushort projectNo, const ushort dateIndex, const GraphType graphType ) const
{	switch( graphType )
	{	case HostAvg:   return m_projectData[ projectNo ].m_projectDateData[ dateIndex ].m_toolTipHostAvg;
		case HostTotal: return m_projectData[ projectNo ].m_projectDateData[ dateIndex ].m_toolTipHostTotal;
		case UserAvg:   return m_projectData[ projectNo ].m_projectDateData[ dateIndex ].m_toolTipUserAvg;
		case UserTotal: return m_projectData[ projectNo ].m_projectDateData[ dateIndex ].m_toolTipUserTotal;
	}

	Q_ASSERT( false );
	return 0;
}

//	valid
//	=====

bool StatisticsData::valid( const ushort projectNo, const ushort i ) const
{	return m_projectData[ projectNo ].m_projectDateData[ i ].m_valid;
}

//	value
//	=====

double StatisticsData::value( const ushort projectNo, const ushort dateIndex, const GraphType graphType ) const
{	switch( graphType )
	{	case HostAvg:   return m_projectData[ projectNo ].m_projectDateData[ dateIndex ].m_hostAvg;
		case HostTotal: return m_projectData[ projectNo ].m_projectDateData[ dateIndex ].m_hostTotal;
		case UserAvg:   return m_projectData[ projectNo ].m_projectDateData[ dateIndex ].m_userAvg;
		case UserTotal: return m_projectData[ projectNo ].m_projectDateData[ dateIndex ].m_userTotal;
	}

	Q_ASSERT( false );
	return 0;
}
