/***************************************************************************
 *                                                                         *
 *   Copyright (C) 2006 by Kevin Gilbert                                   *
 *   kev.gilbert@cdu.edu.au                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>

#include <qdatetime.h>
#include <qevent.h>
#include <qpainter.h>
#include <qpopupmenu.h>
#include <qstringlist.h>

#include <kdebug.h>
#include <kglobal.h>
#include <kglobalsettings.h>
#include <klocale.h>

#include "gui_rpc_client.h"
#include "statisticscontroldata.h"
#include "statisticsdata.h"
#include "statisticsgraphview.h"
#include "statstooltip.h"

//	constructor
//	===========

StatisticsGraphView::StatisticsGraphView( const CPtr client, QWidget* parent, const char* name )
: QWidget( parent, name ),
  Columns( 10 ),
  Margin( 10 ),
  PointSize( 8 ),
  Rows( 10 ),
  m_client( client ),
  m_graphType( GraphType( uint( -1 ))),	// this will be set to a real value via a xfer from the control widget
  m_labelsCredit( NULL ),
  m_labelsDate( NULL ),
  m_noOfProjects( 0 ),
  m_statsData( NULL ),
  m_toolTip( new StatsToolTip( this ))
{	setMouseTracking( true );
}

//	destructor
//	==========

StatisticsGraphView::~StatisticsGraphView( )
{	if( m_labelsCredit != NULL )
		delete [] m_labelsCredit;

	if( m_labelsDate != NULL )
		delete [] m_labelsDate;

	if( m_statsData != NULL )
		delete m_statsData;

	delete m_toolTip;
}

//	buildLabels
//	===========

void StatisticsGraphView::buildLabels( QPainter* painter )
{	if( !m_isNewData )
		return;

	if( m_labelsCredit == NULL )
	{	m_labelsCredit = new Labels[ Rows    + 1 ];
		m_labelsDate   = new Labels[ Columns + 1 ];
	}

	double maxValue    = m_statsData->maxValue( m_projectData->guiRPCIndex( ), m_graphType, m_projectData->stacked( ));
	double minValue    = m_statsData->minValue( m_projectData->guiRPCIndex( ), m_graphType, m_projectData->stacked( ));
	double increment   = (maxValue - minValue) / Rows;
	ushort creditWidth = 0;

	m_xWidth = painter->fontMetrics( ).width( "X" );

	for( ushort i = 0; i < Rows + 1; i++ )
	{	double value = minValue + (Rows - i) * increment;

		m_labelsCredit[ i ].label1.sprintf( "%d", uint( value ));
		m_labelsCredit[ i ].label2 = "";

		ushort widthLabel1 = painter->fontMetrics( ).width( m_labelsCredit[ i ].label1 );
		ushort widthLabel2 = painter->fontMetrics( ).width( m_labelsCredit[ i ].label2 );
		ushort width       = widthLabel1 + widthLabel2;

		if( creditWidth < width )
		{	creditWidth = width;
			m_dotPos    = m_xWidth + widthLabel1;
		}
	}

	m_dotPos  += m_graphRect.left( );

	uint maxDate     = m_statsData->maxDate( );
	uint minDate     = m_statsData->minDate( );
	increment        = (maxDate - minDate) / Columns;
	ushort dateWidth = 0;
	QDateTime date;

	for( ushort i = 0; i < Columns + 1; i++ )
	{	date.setTime_t( uint( minDate + i * increment ));
		QStringList list = QStringList::split( ' ', KGlobal::locale( )->formatDateTime( date, true, true ));
		m_labelsDate[ i ].label1 = list[ 0 ];
		m_labelsDate[ i ].label2 = list[ 1 ];

		ushort widthLabel1 = painter->fontMetrics( ).width( m_labelsDate[ i ].label1 );
		ushort widthLabel2 = painter->fontMetrics( ).width( m_labelsDate[ i ].label2 );

		dateWidth = _max_( dateWidth, _max_( widthLabel1, widthLabel2 ));
	}

	m_graphRect.setBottom( m_graphRect.bottom( ) - 2 * m_xWidth - dateWidth );
	m_graphRect.setLeft( m_graphRect.left( ) + creditWidth + 2 * m_xWidth );
	m_graphRect.setRight( m_graphRect.right( ) - painter->fontMetrics( ).height( ) - m_xWidth );
	m_graphRect.setTop( m_graphRect.top( ) + painter->fontMetrics( ).height( ));
}

//	createAndDrawPointRect
//	======================

void StatisticsGraphView::createAndDrawPointRect( const QPoint& point,
												  const ushort guiRPCIndex,
												  const ushort i,
												  QPainter* painter )
{	if( m_isNewData )
	{	QRect pointRect( point.x( ) - PointSize / 2, point.y( ) - PointSize / 2, PointSize, PointSize );
		m_statsData->point( guiRPCIndex, i, m_graphType, pointRect );
	}

	painter->fillRect( m_statsData->point( guiRPCIndex, i, m_graphType ),
					   m_statsData->interpolated( guiRPCIndex, i ) ? KGlobalSettings::highlightedTextColor( ) :
																	 KGlobalSettings::buttonTextColor( ));
}

//	drawGrid
//	========

void StatisticsGraphView::drawGrid( QPainter* painter )
{	painter->save( );

	QFont  font     = painter->font( );
	short  fontSize = m_graphRect.height( ) / 2;
	QRect  nameRect;

	while( fontSize >= 8 )
	{	font.setPixelSize( fontSize );
		painter->setFont( font );
		nameRect = painter->fontMetrics( ).boundingRect( m_projectName );

		if( nameRect.width( ) < m_graphRect.width( ))
			break;

		fontSize = ushort( 0.9 * fontSize );
	}

	if( fontSize >= 8 )
	{	ushort x = m_graphRect.left( ) + (m_graphRect.width( )  - nameRect.width(  )) / 2;
		ushort y = m_graphRect.top( )  + (m_graphRect.height( ) - nameRect.height( )) / 2 - painter->fontMetrics( ).descent();

		nameRect.moveTopLeft( QPoint( x, y ));

		painter->setPen( KGlobalSettings::baseColor( ));
		painter->drawText( nameRect, Qt::AlignCenter | Qt::AlignTop | Qt::DontClip, m_projectName );
	}

	painter->restore( );
	painter->save( );

	painter->setPen( KGlobalSettings::textColor( ));
	double interval = m_graphRect.width( ) / double( Rows );

	for( ushort i = 0; i < Rows + 1; i++ )
	{	ushort x = ushort( m_graphRect.left( ) + i * interval );
		painter->drawLine( x, m_graphRect.top( ), x, m_graphRect.bottom( ));
	}

	interval = m_graphRect.height( ) / double( Columns );

	for( ushort i = 0; i < Columns + 1; i++ )
	{	ushort y = ushort( m_graphRect.top( ) + i * interval );
		painter->drawLine( m_graphRect.left( ), y, m_graphRect.right( ), y );
	}

	for( ushort i = 0; i < Rows + 1; i++ )
	{	QRect  lRect = painter->fontMetrics( ).boundingRect( m_labelsCredit[ i ].label1 );
		ushort x     = m_dotPos - lRect.width( );
		ushort y     = ushort( m_graphRect.top( ) + i * interval + lRect.height( ) / 2 );

		painter->drawText( x, y, m_labelsCredit[ i ].label1 );
	}

	interval = m_graphRect.width( ) / double( Rows );
	painter->rotate( 90 );

	for( ushort i = 0; i < Columns + 1; i++ )
	{	QRect  rRect = painter->fontMetrics( ).boundingRect( m_labelsDate[ i ].label1 );
		ushort x     = ushort( m_graphRect.left( ) + i * interval + painter->fontMetrics( ).height( ) / 8 );
		ushort y     = m_graphRect.bottom( ) + m_xWidth;
		painter->drawText( y, -x, m_labelsDate[ i ].label1 );
		painter->drawText( y, -x + rRect.height( ) + 2, m_labelsDate[ i ].label2 );
	}

	painter->restore( );
}

//	drawNoData
//	==========

void StatisticsGraphView::drawNoData( QPainter* painter )
{	QFont fontSave  = painter->font( );
	QFont fontTitle = fontSave;

	fontTitle.setBold( true );
	fontTitle.setPointSize( 20 );
	fontTitle.setUnderline( true );

	painter->setFont( fontTitle );
	painter->setPen( KGlobalSettings::highlightColor( ));

	QString text = "No statistical data available";
	ushort  x    = _max_( 0, (m_graphRect.width( ) - painter->fontMetrics( ).width( text )) / 2 );
	ushort  y    = m_graphRect.top( ) + _max_( 0, (m_graphRect.height( ) - painter->fontMetrics( ).height( )) / 2 );

	painter->drawText( x, y, text );
	painter->setFont( fontSave );
}

//	drawProjectPlot
//	===============

void StatisticsGraphView::drawProjectPlot( const ushort guiRPCIndex, QPainter* painter )
{	ushort i;

	for( i = 0; i < m_statsData->noOfDataPoints( ); i++ )
		if( m_statsData->valid( guiRPCIndex, i ))
			break;

	if( i >= m_statsData->noOfDataPoints( ))
		return;

	painter->save( );

	uint   maxDate     = m_statsData->maxDate( );
	uint   minDate     = m_statsData->minDate( );
	double xInt        = double( m_graphRect.width( )) / (maxDate - minDate);
	double scaledValue = m_statsData->scaledValue(  guiRPCIndex, i, m_graphType );
	ushort x           = m_graphRect.left( ) + ushort( xInt * (m_statsData->date(  guiRPCIndex, i ) - minDate));
	ushort y           = ushort( m_graphRect.bottom( ) - scaledValue * m_graphRect.height( ));
	QPoint from( x, y );

	createAndDrawPointRect( from, guiRPCIndex, i, painter );

	for( i++; i < m_statsData->noOfDataPoints( ); i++ )
	{	if( !m_statsData->valid( guiRPCIndex, i ))
			continue;

		scaledValue = m_statsData->scaledValue(  guiRPCIndex, i, m_graphType );
		x = m_graphRect.left( ) + ushort( xInt * (m_statsData->date(  guiRPCIndex, i ) - minDate));
		y = ushort( m_graphRect.bottom( ) - scaledValue * m_graphRect.height( ));
		QPoint to( x, y );

		createAndDrawPointRect( to, guiRPCIndex, i, painter );
		painter->drawLine( from, to );
		from = to;
	}

	painter->restore( );
}

//	drawTitle
//	=========

ushort StatisticsGraphView::drawTitle( QPainter* painter, const QString& title )
{	QFont fontSave  = painter->font( );
	QFont fontTitle = fontSave;
	QPen  penSave   = painter->pen( );

	fontTitle.setBold( true );
	fontTitle.setPointSize( 20 );
	fontTitle.setUnderline( true );

	painter->setFont( fontTitle );
	painter->setPen( KGlobalSettings::highlightedTextColor( ));

	ushort x = (width( ) - painter->fontMetrics( ).width( title )) / 2;
	ushort y = ushort( 1.5 * fontMetrics( ).height( ));

	painter->drawText( x, y, title );
	painter->setFont( fontSave );
	painter->setPen( penSave );

	return 2 * fontMetrics( ).height( );
}

//	findProjectData
//	===============

void StatisticsGraphView::findProjectData( )
{	for( m_projectData = m_projectsData->first( ); m_projectData != NULL; m_projectData = m_projectsData->next( ))
		if( m_projectName == m_projectData->projectName( ))
			return;
}

//	findStatsData
//	=============

bool StatisticsGraphView::findStatsData( const QPoint& pos, const ushort guiRPCIndex, ushort& i )
{	for( i = 0; i < m_statsData->noOfDataPoints( ); i++ )
		if( m_statsData->point( guiRPCIndex, i, m_graphType ).contains( pos ))
			return true;

	return false;
}

//	mouseMoveEvent
//	==============

void StatisticsGraphView::mouseMoveEvent( QMouseEvent* event )
{	if( m_statsData == NULL )
	{	m_toolTip->show( false );
		return;
	}

	if( !m_projectData->stacked( ))
		mouseMoveEvent( event->pos( ), m_projectData->guiRPCIndex( ));
	else
	for( SCDPtr projectData = m_projectsData->first( ); projectData != NULL; projectData = m_projectsData->next( ))
		if( mouseMoveEvent( event->pos( ), projectData->guiRPCIndex( )))
			break;
}

//	mouseMoveEvent
//	==============

bool StatisticsGraphView::mouseMoveEvent( const QPoint& pos, const ushort guiRPCIndex )
{	ushort i;

	if( !findStatsData( pos, guiRPCIndex, i ))
	{	m_toolTip->show( false );
		return false;
	}

	m_toolTip->setTipStuff( m_statsData->point(  guiRPCIndex, i, m_graphType ),
							m_statsData->toolTipText(  guiRPCIndex, i, m_graphType ));
	m_toolTip->show( true );
	return true;
}

//	mouseReleaseEvent
//	=================

void StatisticsGraphView::mouseReleaseEvent( QMouseEvent* event )
{	if( !(event->state( ) & Qt::RightButton) || (event->state( ) & (Qt::ShiftButton | Qt::ControlButton | Qt::AltButton)) )
		return;

	if( !m_projectData->stacked( ))
		mouseReleaseEvent( event, m_projectData->guiRPCIndex( ));
	else
	for( SCDPtr projectData = m_projectsData->first( ); projectData != NULL; projectData = m_projectsData->next( ))
		if( mouseReleaseEvent( event, projectData->guiRPCIndex( )))
			break;
}

//	mouseReleaseEvent
//	=================

bool StatisticsGraphView::mouseReleaseEvent( QMouseEvent* event, const ushort guiRPCIndex )
{	if( !findStatsData( event->pos( ), guiRPCIndex, m_removeIndex ))
	{	m_toolTip->show( false );
		return false;
	}

	QPopupMenu* menu = new QPopupMenu( this );
	menu->insertItem( i18n( "&Remove data point" ), this, SLOT( slotRemoveDataPoint( )));
	menu->exec( event->globalPos( ));

	m_removeGUIRPCIndex = guiRPCIndex;
	return true;
}

//	paintEvent
//	==========

void StatisticsGraphView::paintEvent( QPaintEvent* /* event */)
{	QString title;

	switch( m_graphType )
	{	case HostAvg:   title = "Host Average"; break;
		case HostTotal: title = "Host Total";   break;
		case UserAvg:   title = "User Average"; break;
		case UserTotal: title = "User Total";   break;
	}

	if( m_removedDataPoints.count( ))
		title += "*";

	QPainter painter( this );
	ushort top = drawTitle( &painter, title );

	if( m_isNewData )
		m_graphRect = QRect( Margin, top, width( ) - 2 * Margin, height( ) - top - Margin );

	painter.fillRect( QRect( Margin, top, width( ) - 2 * Margin, height( ) - top - Margin ),
					  KGlobalSettings::alternateBackgroundColor( ));

	if( !m_noOfProjects )
	{	drawNoData( &painter );
		return;
	}

	painter.save( );
	buildLabels( &painter );
	drawGrid( &painter );

	painter.setPen( KGlobalSettings::highlightedTextColor( ));
	painter.setClipRect( m_graphRect );

	if( !m_projectData->stacked( ))
		drawProjectPlot( m_projectData->guiRPCIndex( ), &painter );
	else
	for( SCDPtr projectData = m_projectsData->first( ); projectData != NULL; projectData = m_projectsData->next( ))
		drawProjectPlot( projectData->guiRPCIndex( ), &painter );

	painter.restore( );
	m_isNewData = false;
}

//	resizeEvent
//	===========

void StatisticsGraphView::resizeEvent( QResizeEvent* /* event */)
{	m_isNewData = true;
}

//	setInitialSh1t
//	==============

void StatisticsGraphView::setInitialSh1t( QPtrList<StatisticsControlData>* projectsData,
										  const GraphType graphType,
										  const QString&  projectName )
{	m_graphType    = graphType;
	m_projectsData = projectsData;
	m_projectName  = projectName;

	findProjectData( );
}

//	slotHostAverage
//	===============

void StatisticsGraphView::slotHostAverage( const QString& projectName )
{	m_graphType   = HostAvg;
	m_projectName = projectName;

	update( true );
}

//	slotHostTotal
//	=============

void StatisticsGraphView::slotHostTotal( const QString& projectName )
{	m_graphType   = HostTotal;
	m_projectName = projectName;

	update( true );
}

//	slotRemoveDataPoint
//	===================

void StatisticsGraphView::slotRemoveDataPoint( )
{	m_removedDataPoints.append( m_statsData->date( m_removeGUIRPCIndex, m_removeIndex ));
	update( true );
}

//	slotUserAverage
//	===============

void StatisticsGraphView::slotUserAverage( const QString& projectName )
{	m_graphType   = UserAvg;
	m_projectName = projectName;

	update( true );
}

//	slotUserTotal
//	=============

void StatisticsGraphView::slotUserTotal( const QString& projectName )
{	m_graphType   = UserTotal;
	m_projectName = projectName;

	update( true );
}

//	update
//	======

bool StatisticsGraphView::update( const bool forceUpdate )
{	PROJECTS projects;
	int statr = m_client->rpcClient.get_statistics( projects );

	if( statr )
	{	m_client->reason.sprintf( i18n( "get_statistics returned error %d" ), statr );
		return false;
	}

	if( forceUpdate )
		findProjectData( );
	else
	if( m_statsData != NULL && m_statsData->sameData( projects ))
		return true;

	m_isNewData = true;
	m_noOfProjects = projects.projects.size( );

	if( !m_noOfProjects )
	{	if( forceUpdate )
			QWidget::update( );

		return true;
	}

	if( m_statsData != NULL )
		delete m_statsData;

	m_statsData = new StatisticsData( projects,
									  m_projectsData,
									  m_client->clientName,
									  m_projectData->stacked( ),
									  &m_removedDataPoints );
	QWidget::update( );
	return true;
}
