/***************************************************************************
 *                                                                         *
 *   Copyright (C) 2006 by Kevin Gilbert                                   *
 *   kev.gilbert@cdu.edu.au                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#include <qdatetime.h>
#include <kdebug.h>
#include <kglobal.h>
#include <klocale.h>
#include <kmessagebox.h>

#include "authenticate.h"
#include "gui_rpc_client.h"
#include "mainwidgetview.h"
#include "transferslistview.h"
#include "transfersviewitem.h"

//	constructor
//	===========

TransfersListView::TransfersListView( const CPtr client, QWidget* parent, const char* name )
: KListView( parent, name ),
  m_client( client ),
  m_waitingOnAutoAuth( false )
{	setAllColumnsShowFocus( true );
	setSelectionMode( QListView::Extended );
	setShowSortIndicator( true );

	addColumn( "Project"       );
	addColumn( "File"          );
	addColumn( "Size (KB)"     );
	addColumn( "Elapsed time"  );
	addColumn( "Status"        );
	addColumn( "First attempt" );
	addColumn( "Retries"       );
	addColumn( "Progress"      );
	addColumn( "Speed (KBps)"  );
	addColumn( "ETC"           );

	setColumnAlignment( 2, Qt::AlignRight );
	setColumnAlignment( 3, Qt::AlignRight );
	setColumnAlignment( 5, Qt::AlignRight );
	setColumnAlignment( 6, Qt::AlignRight );
	setColumnAlignment( 7, Qt::AlignRight );
	setColumnAlignment( 8, Qt::AlignRight );
	setColumnAlignment( 9, Qt::AlignRight );

	for( byte i = 0; i < columns( ); i++ )
		setColumnWidthMode( i, QListView::Manual );

	connect( this, SIGNAL( selectionChanged( )), SLOT( slotSelectionChanged( )));
}

//	abortOrRetryTransfer
//	====================

void TransfersListView::abortOrRetryTransfer( const Operation op, const QString& name )
{	FILE_TRANSFERS transfers;
	int statr = m_client->rpcClient.get_file_transfers( transfers );

	if( statr )
	{	kdDebug( ) << "TransfersListView::retryTransfer - get_file_transfers returned error " << statr << endl;
		return;
	}

	uint i;

	for( i = 0; i < transfers.file_transfers.size( ); i++ )
		if( transfers.file_transfers[ i ]->name.c_str( ) == name )
			break;

	if( i >= transfers.file_transfers.size( ))
	{	kdDebug( ) << "TransfersListView::retryTransfer - transfer >" << name << "< no longer current" << endl;
		return;
	}

	QString textOp;

	switch( op )
	{	case Abort: textOp = "abort"; break;
		case Retry: textOp = "retry"; break;

		default:
			Q_ASSERT( false );
			return;
	}

	statr = m_client->rpcClient.file_transfer_op( *transfers.file_transfers[ i ], textOp );

	if( statr )
	{	kdDebug( ) << "TransfersListView::retryTransfer - file_transfer_op returned error " << statr << endl;
		return;
	}
}

//	authenticationFinished
//	======================

void TransfersListView::authenticationFinished( const CPtr /*client */, const bool successful)
{	if( !successful || !m_waitingOnAutoAuth )
		return;

	TransfersViewItem* item;

	for( item = m_transfers.first( ); item != NULL; item = m_transfers.next( ))
	{	kdDebug( ) << "automatically retrying >" << item->key( ) << "<\n";
		abortOrRetryTransfer( Retry, item->key( ));
	}
}

//	processNewTransfers
//	===================

void TransfersListView::processNewTransfers( QPtrList<FILE_TRANSFER>& newTransferList )
{	bool autoAuthEmitted = false;
	FILE_TRANSFER* transfer;

	for( transfer = newTransferList.first( ); transfer != NULL; transfer = newTransferList.first( ))
	{	newTransferList.removeFirst( );
		TransfersViewItem* item = new TransfersViewItem( this, transfer );
		m_transfers.append( item );
		update( item, transfer );

		if( transfer->status && m_client->authenticate->authReqd( ) && !autoAuthEmitted )	// status != downloading
		{	autoAuthEmitted = true;

			if( !m_client->authenticate->autoAuthenticate( ))
				m_waitingOnAutoAuth = false;
			else
				m_waitingOnAutoAuth = m_client->authenticate->authenticate( );
		}
	}

	m_transfers.sort( );
}

//	slotAbortClicked
//	================

void TransfersListView::slotAbortClicked( )
{	TransfersViewItem* item;

	for( item = m_transfers.first( ); item != NULL; item = m_transfers.next( ))
		if( item->isSelected( ) && item->isRetryCandidate( ))
		{	if( KMessageBox::Yes == KMessageBox::questionYesNo( this, i18n( "Are you sure you want to abort this transfer. If you do, you wont get any credit for it - and it cannot be recovered!" ), i18n( "Confirm deletion" )))
				abortOrRetryTransfer( Abort, item->key( ));
		}
}

//	slotRetryClicked
//	================

void TransfersListView::slotRetryClicked( )
{	TransfersViewItem* item;

	for( item = m_transfers.first( ); item != NULL; item = m_transfers.next( ))
		if( item->isSelected( ) && item->isRetryCandidate( ))
			abortOrRetryTransfer( Retry, item->key( ));
}

//	slotSelectionChanged
//	====================

void TransfersListView::slotSelectionChanged( )
{	bool enableAbort = false;
	bool enableRetry = false;
	TransfersViewItem* item;

	for( item = m_transfers.first( ); item != NULL; item = m_transfers.next( ))
		if( item->isSelected( ) && item->isAbortCandidate( ))
		{	enableAbort = true;
			break;
		}

	for( item = m_transfers.first( ); item != NULL; item = m_transfers.next( ))
		if( item->isSelected( ) && item->isRetryCandidate( ))
		{	enableRetry = true;
			break;
		}

	emit( enableButtons( enableAbort, enableRetry ));
}

//	update
//	======

bool TransfersListView::update( )
{	FILE_TRANSFERS transfers;
	int statr = m_client->rpcClient.get_file_transfers( transfers );

	if( statr )
	{	m_client->reason.sprintf( i18n( "get_file_transfers returned error %d" ), statr );
		return false;
	}

	for( TransfersViewItem* item = m_transfers.first( ); item != NULL; item = m_transfers.next( ))
		item->updated( false );

	QPtrList<FILE_TRANSFER> newTransferList;

	for( uint i = 0; i < transfers.file_transfers.size( ); i++ )
	{	TransfersViewItem* item;

		for( item = m_transfers.first( ); item != NULL; item = m_transfers.next( ))
		{	statr = QString::compare( item->key( ), transfers.file_transfers[ i ]->name.c_str( ));

			if( !statr )
				break;

			if( statr > 0 )
			{	item = NULL;
				break;
			}
		}

		if( item == NULL )
			newTransferList.append( transfers.file_transfers[ i ]);
		else
		{	update( item, transfers.file_transfers[ i ]);
			item->updated( true );
		}
	}

	for( TransfersViewItem* item = m_transfers.last( ); item != NULL; item = m_transfers.prev( ))
		if( !item->updated( ))
		{	m_transfers.remove( );
			delete item;
		}

	if( newTransferList.count( ))
		processNewTransfers( newTransferList );

	return true;
}

//	update
//	======

void TransfersListView::update( TransfersViewItem* item, FILE_TRANSFER* transfer )
{	PROJECT* project = m_client->state.lookup_project( transfer->project_url );

	if( project == NULL && MainWidgetView::getState( m_client ))
		project = m_client->state.lookup_project( transfer->project_url );

	item->setText( 0, project == NULL ? "**** unknown ****" : project->project_name.c_str( ));
	item->setText( 1, transfer->name.c_str( ));

	QString text;
	text.sprintf( "%.1f", transfer->nbytes / 1024 );
	item->setText( 2, text );
	item->setText( 3, MainWidgetView::secondsToString( transfer->time_so_far ));

	text.sprintf( "%.1f%%", 100 * transfer->bytes_xferred / transfer->nbytes );
	item->setText( 7, text );

	text.sprintf( "%.1f", transfer->xfer_speed / 1024 );
	item->setText( 8, text );

	item->setText( 9, transfer->xfer_speed ? MainWidgetView::secondsToString( (transfer->nbytes - transfer->bytes_xferred) / transfer->xfer_speed ) : "?" );

	uint now = QDateTime::currentDateTime( ).toTime_t( );
	QDateTime firstRequestTime;

	switch( transfer->status )
	{	case 0:
			item->setAbortCandidate( true );
			item->setRetryCandidate( false );
			item->setText( 4, "Downloading" );
			break;

		case 1:
			firstRequestTime.setTime_t( transfer->first_request_time );

			item->setAbortCandidate( true );
			item->setRetryCandidate( true );

			item->setText( 4, "Retry in " + MainWidgetView::secondsToString( transfer->next_request_time - now ));
			item->setText( 5, KGlobal::locale( )->formatDateTime( firstRequestTime, true, true ));
			item->setText( 6, QString::number( transfer->num_retries ));
			break;

		default:
			text.sprintf( "**** unknown (%d) ****", transfer->status );

			item->setAbortCandidate( true );
			item->setRetryCandidate( false );
			item->setText( 4, text );
	}
}
